# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gc
import os
import sys
import time
from collections import deque
from datetime import datetime, timedelta
from typing import Any, Callable, Optional, Union

import torch
import torch.profiler
from megatron.core.distributed import DistributedDataParallel as DDP
from megatron.core.distributed.fsdp.mcore_fsdp_adapter import FullyShardedDataParallel as megatron_FSDP
from megatron.core.full_cuda_graph import FullCudaGraphWrapper
from megatron.core.num_microbatches_calculator import (
    get_current_global_batch_size,
    get_current_running_global_batch_size,
    get_num_microbatches,
    update_num_microbatches,
)
from megatron.core.optimizer import MegatronOptimizer
from megatron.core.optimizer.distrib_optimizer import DistributedOptimizer
from megatron.core.optimizer.qk_clip import clip_qk
from megatron.core.optimizer_param_scheduler import OptimizerParamScheduler
from megatron.core.parallel_state import update_pg_timeout
from megatron.core.pipeline_parallel.p2p_communication import P2PCommunicator
from megatron.core.pipeline_parallel.schedules import get_forward_backward_func
from megatron.core.pipeline_parallel.utils import (
    is_pp_first_stage,
    is_pp_last_stage,
)
from megatron.core.process_groups_config import ProcessGroupCollection
from megatron.core.rerun_state_machine import RerunDataIterator, get_rerun_state_machine
from megatron.core.transformer import MegatronModule
from megatron.core.transformer.cuda_graphs import TECudaGraphHelper
from megatron.core.transformer.enums import CudaGraphScope
from megatron.core.utils import check_param_hashes_across_dp_replicas, get_model_config
from modelopt.torch.distill.plugins.megatron import get_tensor_shapes_adjust_fn_for_distillation

from megatron.bridge.training import fault_tolerance
from megatron.bridge.training.callbacks import CallbackContext, CallbackManager, should_fire
from megatron.bridge.training.checkpointing import maybe_finalize_async_save, save_checkpoint
from megatron.bridge.training.config import ConfigContainer
from megatron.bridge.training.eval import evaluate_and_print_results
from megatron.bridge.training.forward_step_func_types import ForwardStepCallable
from megatron.bridge.training.initialize import destroy_global_state
from megatron.bridge.training.nvrx_straggler import (
    check_nvrx_straggler_detection,
    safe_shutdown_nvrx_straggler_manager,
)
from megatron.bridge.training.profiling import (
    handle_profiling_step,
    handle_profiling_stop,
    initialize_pytorch_profiler,
    should_profile_rank,
)
from megatron.bridge.training.state import GlobalState
from megatron.bridge.training.tensor_inspect import (
    tensor_inspect_end_if_enabled,
    tensor_inspect_step_if_enabled,
)
from megatron.bridge.training.utils import flop_utils
from megatron.bridge.training.utils.log_utils import append_to_progress_log, barrier_and_log
from megatron.bridge.training.utils.train_utils import (
    calc_params_l2_norm,
    logical_and_across_model_parallel_group,
    prepare_forward_step_func,
    reduce_max_stat_across_model_parallel_group,
    training_log,
)
from megatron.bridge.utils.common_utils import get_world_size_safe, print_rank_0


def train(
    forward_step_func: ForwardStepCallable,
    model: list[MegatronModule],
    optimizer: MegatronOptimizer,
    scheduler: OptimizerParamScheduler,
    train_data_iterator: Optional[Union[RerunDataIterator, list[RerunDataIterator]]],
    valid_data_iterator: Optional[Union[RerunDataIterator, list[RerunDataIterator]]],
    global_state: GlobalState,
    checkpointing_context: dict[str, Any],
    pg_collection: ProcessGroupCollection,
    process_non_loss_data_func: Optional[Callable] = None,
    non_loss_data_func: Optional[Callable] = None,
    callback_manager: CallbackManager | None = None,
) -> None:
    """Main training loop.

    Handles the overall training process, including the iteration loop,
    calling train_step, evaluation, checkpointing, logging, and exit conditions.

    Args:
        forward_step_func: Callable that executes a single forward step.
        model: list of model chunks (potentially wrapped in DDP).
        optimizer: The optimizer instance.
        scheduler: The learning rate scheduler instance.
        train_data_iterator: Iterator for the training dataset.
        valid_data_iterator: Iterator for the validation dataset.
        global_state: The GlobalState object holding various training states.
        checkpointing_context: Context dictionary for checkpointing.
        process_non_loss_data_func: Optional function to process non-loss data during evaluation.
        non_loss_data_func: Optional function to compute non-loss data during evaluation.
        callback_manager: Optional CallbackManager for custom callback execution.

    Warnings:
        This is an experimental API and is subject to change in backwards
        incompatible ways without notice.
    """
    config: ConfigContainer = global_state.cfg
    model_config = get_model_config(model[0])
    train_config = config.train
    timers = global_state.timers
    straggler_timer = global_state.straggler_timer
    energy_monitor = global_state.energy_monitor

    # Prepare forward_step_func (check signature and inject state if needed).
    # This is done once to prevent creating new partial objects every iteration.
    #
    # Note on reference semantics:
    # - functools.partial stores a reference to global_state, not a copy
    # - When global_state.train_state.step changes, the partial sees the updated value
    # - This is safe because GlobalState is a mutable object passed by reference
    #
    # For functors (classes with __call__ defined):
    # - For functors: partial(functor_instance, state) still allows functor's internal state to work
    # - inspect.signature() properly inspects the __call__ method of functors
    wrapped_forward_step_func = prepare_forward_step_func(forward_step_func, global_state)

    # Turn on training mode which enables dropout.
    for model_module in model:
        model_module.train()

    # Tracking loss.
    total_loss_dict = {}

    # Make sure rerun_state_machine has the right iteration loaded from checkpoint.
    rerun_state_machine = get_rerun_state_machine()
    if rerun_state_machine.current_iteration != global_state.train_state.step:
        print_rank_0(f"Setting rerun_state_machine.current_iteration to {global_state.train_state.step}...")
        rerun_state_machine.current_iteration = global_state.train_state.step

    num_floating_point_operations_so_far = global_state.train_state.floating_point_operations_so_far
    num_floating_point_operations_since_last_log_event = 0.0

    if energy_monitor is not None:
        energy_monitor.setup()
        energy_monitor.resume()

    timers("interval-time", log_level=0).start(barrier=True)
    report_memory_flag = True
    pre_hook_enabled = False
    should_exit = False
    exit_code = 0

    if train_config.manual_gc:
        # Disable the default garbage collector and perform the collection manually.
        # This is to align the timing of garbage collection across ranks.
        assert train_config.manual_gc_interval >= 0, (
            "Manual garbage collection interval should be larger than or equal to 0"
        )
        gc.disable()
        gc.collect()

    if config.straggler and config.straggler.log_straggler:
        world = torch.distributed.get_world_size()
        rank = torch.distributed.get_rank()
        mmcnt = config.straggler.straggler_minmax_count
        straggler_timer.configure(
            world,
            rank,
            mmcnt=mmcnt,
            enabled=not config.straggler.disable_straggler_on_startup,
            port=config.straggler.straggler_ctrlr_port,
        )

    # Initialize NVRx straggler detection if enabled
    nvrx_straggler_manager = global_state.nvrx_straggler_manager
    wrapped_train_step = train_step  # Default to original function
    if nvrx_straggler_manager is not None:
        try:
            # Initialize the straggler detector first
            nvrx_straggler_manager.initialize()
            # Wrap the train_step function for monitoring
            # The wrapped function must be used instead of the original to collect profiling data
            wrapped_train_step = nvrx_straggler_manager.wrap_train_step_function(train_step)
        except Exception as e:
            print_rank_0(f"Failed to initialize NVRx straggler detection: {e}")
            # Set to None to disable further checks
            global_state._nvrx_straggler_manager = None

    num_microbatches = get_num_microbatches()
    eval_duration = 0.0
    eval_iterations = 0

    prof = None
    nsys_nvtx_context = None  # NVTX context for nsys profiling
    prof_config = config.profiling
    if prof_config and should_profile_rank(prof_config, torch.distributed.get_rank()):
        if prof_config.use_pytorch_profiler:
            prof = initialize_pytorch_profiler(prof_config, config.logger.tensorboard_dir)
            prof.start()

    # Megatron FSDP and FSDP2 does not have this hook
    should_toggle_forward_pre_hook = should_disable_forward_pre_hook(
        config.ddp.use_megatron_fsdp,
        config.optimizer.use_distributed_optimizer,
        config.ddp.overlap_param_gather,
    )
    # Disable forward pre-hook to start training to ensure that errors in checkpoint loading
    # or random initialization don't propagate to all ranks in first all-gather (which is a
    # no-op if things work correctly).
    if should_toggle_forward_pre_hook:
        disable_forward_pre_hook(model, param_sync=False)
        # Also remove param_sync_func temporarily so that sync calls made in
        # `forward_backward_func` are no-ops.
        param_sync_func = model_config.param_sync_func
        model_config.param_sync_func = None
        pre_hook_enabled = False
    # Also, check weight hash across DP replicas to be very pedantic.
    if train_config.check_weight_hash_across_dp_replicas_interval is not None:
        assert check_param_hashes_across_dp_replicas(model, cross_check=True), (
            "Parameter hashes not matching across DP replicas"
        )
        torch.distributed.barrier()
        print_rank_0(f">>> Weight hashes match after {global_state.train_state.step} iterations...")

    # Capture CUDA Graphs.
    cuda_graph_helper = None
    if model_config.cuda_graph_impl == "transformer_engine":
        cuda_graph_helper = TECudaGraphHelper(
            model=model,
            config=model_config,
            seq_length=config.model.seq_length,
            micro_batch_size=config.train.micro_batch_size,
            optimizers=[optimizer],
        )

    # Track train step elapsed time for throughput logging
    history_wct = None
    if config.logger.log_throughput_to_tensorboard:
        history_wct = deque(maxlen=config.logger.throughput_window_size + 1)

    # Wrap forward_backward_func for Full iteration CUDA graph
    forward_backward_func = get_forward_backward_func(
        pp_size=pg_collection.pp.size(),
        vp_size=config.model.virtual_pipeline_model_parallel_size,
    )
    if config.model.cuda_graph_impl == "local" and CudaGraphScope.full_iteration in config.model.cuda_graph_scope:
        forward_backward_func = FullCudaGraphWrapper(
            forward_backward_func, cuda_graph_warmup_steps=config.model.cuda_graph_warmup_steps
        )

    start_iteration = global_state.train_state.step
    print_rank_0(f"Starting training loop at iteration {start_iteration}")

    if should_fire(callback_manager, "on_train_start"):
        callback_manager.fire(
            "on_train_start",
            CallbackContext(
                state=global_state,
                model=model,
                user_state=callback_manager.user_state,
                optimizer=optimizer,
                scheduler=scheduler,
            ),
        )

    # Run training iterations till done.
    while global_state.train_state.step < train_config.train_iters:
        # Handle profiling for this step
        nvtx_ctx = handle_profiling_step(
            prof_config,
            global_state.train_state.step,
            torch.distributed.get_rank(),
            prof,
        )
        if nvtx_ctx is not None:
            nsys_nvtx_context = nvtx_ctx

        fault_tolerance.on_checkpointing_start(global_state)
        maybe_finalize_async_save(global_state=global_state, ckpt_cfg=config.checkpoint, blocking=False)
        fault_tolerance.on_checkpointing_end(global_state=global_state, is_async_finalization=True)

        # Update the timeout for all process groups after initialization
        # We update the timeout after the first successful iteration,
        # which takes longer than others usually
        if global_state.train_state.step == start_iteration + 1:
            distributed_timeout_seconds_after_init = global_state.cfg.dist.distributed_timeout_seconds_after_init
            if distributed_timeout_seconds_after_init is not None:
                update_pg_timeout(timedelta(seconds=distributed_timeout_seconds_after_init))

        # Update number of microbatches first without consistency check to decide if a
        # checkpoint should be saved. If the number of microbatches is different
        # from the previous iteration, save a checkpoint. Then run consistency check
        # to make sure training configuration is still valid.
        update_num_microbatches(global_state.train_state.consumed_train_samples, consistency_check=False, verbose=True)
        if get_num_microbatches() != num_microbatches and global_state.train_state.step != 0:
            assert get_num_microbatches() > num_microbatches, (
                f"Number of microbatches should be increasing due to batch size rampup; "
                f"instead going from {num_microbatches} to {get_num_microbatches()}"
            )
            if config.checkpoint.save is not None:
                save_checkpoint_and_time(
                    global_state,
                    model,
                    optimizer,
                    scheduler,
                    num_floating_point_operations_so_far,
                    checkpointing_context,
                    non_persistent_ckpt=False,  # TODO: implement non-persistent checkpointing
                    train_data_iterator=train_data_iterator,
                )
        num_microbatches = get_num_microbatches()
        update_num_microbatches(global_state.train_state.consumed_train_samples, consistency_check=True, verbose=True)

        # Completely skip iteration if needed.
        if _should_skip_and_handle_iteration(global_state, train_data_iterator, pg_collection):
            continue

        # Capture CUDA Graphs after warmup.
        if (
            model_config.cuda_graph_impl == "transformer_engine"
            and cuda_graph_helper is not None
            and not cuda_graph_helper.graphs_created()
            and global_state.train_state.step - start_iteration == model_config.cuda_graph_warmup_steps
        ):
            if model_config.cuda_graph_warmup_steps > 0 and should_toggle_forward_pre_hook:
                disable_forward_pre_hook(model, param_sync=False)
            cuda_graph_helper.create_cudagraphs()
            if model_config.cuda_graph_warmup_steps > 0 and should_toggle_forward_pre_hook:
                enable_forward_pre_hook(model)
                cuda_graph_helper.cuda_graph_set_manual_hooks()

        # Run training step.
        fault_tolerance.on_training_step_start(global_state)

        if should_fire(callback_manager, "on_train_step_start"):
            callback_manager.fire(
                "on_train_step_start",
                CallbackContext(
                    state=global_state,
                    model=model,
                    user_state=callback_manager.user_state,
                    optimizer=optimizer,
                    scheduler=scheduler,
                ),
            )

        (
            loss_dict,
            skipped_iter,
            should_checkpoint,
            should_exit,
            exit_code,
            grad_norm,
            num_zeros_in_grad,
            log_max_attention_logit,
        ) = wrapped_train_step(
            wrapped_forward_step_func,
            train_data_iterator,
            model,
            optimizer,
            scheduler,
            global_state,
            pg_collection,
            forward_backward_func,
        )

        fault_tolerance.on_training_step_end(global_state)

        if should_fire(callback_manager, "on_train_step_end"):
            callback_manager.fire(
                "on_train_step_end",
                CallbackContext(
                    state=global_state,
                    model=model,
                    user_state=callback_manager.user_state,
                    optimizer=optimizer,
                    scheduler=scheduler,
                    loss_dict=loss_dict,
                    grad_norm=grad_norm,
                    skipped_iter=bool(skipped_iter),
                ),
            )

        # Advance NVIDIA DLFw Inspect step if enabled
        tensor_inspect_step_if_enabled(config.tensor_inspect)

        if config.logger.log_throughput_to_tensorboard:
            history_wct.append(time.time() - global_state.start_time)

        if should_checkpoint:
            save_checkpoint_and_time(
                global_state,
                model,
                optimizer,
                scheduler,
                num_floating_point_operations_so_far,
                checkpointing_context,
                train_data_iterator=train_data_iterator,
                non_persistent_ckpt=False,  # TODO: implement non-persistent checkpointing
            )
        if should_exit:
            break

        # Enable forward pre-hooks after first set of forward and backward passes.
        # When running in fp16, skip all NaN iterations until steady-state loss scaling value
        # is reached.
        if global_state.train_state.step == start_iteration:
            if skipped_iter:
                # Only enable forward pre-hook after a training step has successfully run. Relevant
                # for fp16 codepath where first XX iterations are skipped until steady-state loss
                # scale value is reached.
                start_iteration = global_state.train_state.step + 1
            else:
                # Enable forward pre-hook after training step has successfully run. All subsequent
                # forward passes will use the forward pre-hook / `param_sync_func` in
                # `forward_backward_func`.
                if should_toggle_forward_pre_hook:
                    enable_forward_pre_hook(model)
                    model_config.param_sync_func = param_sync_func
                    pre_hook_enabled = True
                    # Set the manual hooks here since it's not set right after the capturing.
                    if (
                        model_config.cuda_graph_impl == "transformer_engine"
                        and model_config.cuda_graph_warmup_steps == 0
                    ):
                        assert cuda_graph_helper.graphs_created(), "CUDA Graphs should have been created."
                        cuda_graph_helper.cuda_graph_set_manual_hooks()

        global_state.train_state.step += 1

        # If fsdp_manual_registration is enabled, manually register FSDP communication buffers after one training step.
        if global_state.train_state.step == start_iteration + 1 and config.ddp.use_megatron_fsdp:
            _maybe_register_fsdp_buffers(config, model)

        dp_size = pg_collection.dp.size()
        batch_size = dp_size * train_config.micro_batch_size * get_num_microbatches()
        global_state.train_state.consumed_train_samples += batch_size
        num_skipped_samples_in_batch = get_current_global_batch_size() - get_current_running_global_batch_size()
        if train_config.decrease_batch_size_if_needed:
            assert num_skipped_samples_in_batch >= 0
        else:
            assert num_skipped_samples_in_batch == 0
        global_state.train_state.skipped_train_samples += num_skipped_samples_in_batch
        num_floating_point_operations_in_batch = flop_utils.num_floating_point_operations(config, batch_size)
        global_state.train_state.floating_point_operations_so_far += num_floating_point_operations_in_batch
        num_floating_point_operations_so_far = global_state.train_state.floating_point_operations_so_far
        num_floating_point_operations_since_last_log_event += num_floating_point_operations_in_batch

        # Logging.
        if hasattr(optimizer, "is_stub_optimizer") and not optimizer.is_stub_optimizer:
            loss_scale = optimizer.get_loss_scale().item()
        else:
            loss_scale = 1.0
        params_norm = None

        if config.logger.log_params_norm:
            params_norm = calc_params_l2_norm(model, model_config, use_megatron_fsdp=config.dist.use_megatron_fsdp)
        learning_rate = None
        decoupled_learning_rate = None
        for param_group in optimizer.param_groups:
            if len(param_group) == 0:
                continue
            if param_group["is_decoupled_lr"]:
                decoupled_learning_rate = param_group["lr"]
            else:
                learning_rate = param_group["lr"]
        report_memory_flag = training_log(
            loss_dict,
            total_loss_dict,
            learning_rate,
            decoupled_learning_rate,
            loss_scale,
            report_memory_flag,
            skipped_iter,
            grad_norm,
            params_norm,
            num_zeros_in_grad,
            config,
            global_state,
            history_wct,
            model,
            log_max_attention_logit,
        )

        if (
            global_state.train_state.do_valid
            and train_config.eval_interval
            and global_state.train_state.step % train_config.eval_interval == 0
        ):
            if energy_monitor is not None:
                energy_monitor.pause()
            timers("interval-time").stop()
            if should_toggle_forward_pre_hook:
                disable_forward_pre_hook(model)
                pre_hook_enabled = False
            if train_config.manual_gc and train_config.manual_gc_eval:
                # Collect all objects.
                gc.collect()
            prefix = f"iteration {global_state.train_state.step}"
            timers("eval-time", log_level=0).start(barrier=True)
            evaluate_and_print_results(
                global_state,
                prefix,
                forward_step_func,
                valid_data_iterator,
                model,
                model_config,
                verbose=False,
                write_to_tensorboard=True,
                process_non_loss_data_func=process_non_loss_data_func,
                non_loss_data_func=non_loss_data_func,
                callback_manager=callback_manager,
            )
            eval_duration += timers("eval-time").elapsed()
            eval_iterations += train_config.eval_iters
            timers("eval-time").stop()

            if train_config.manual_gc and train_config.manual_gc_eval:
                # Collect only the objects created and used in evaluation.
                gc.collect(generation=0)
            if should_toggle_forward_pre_hook:
                enable_forward_pre_hook(model)
                pre_hook_enabled = True
            timers("interval-time", log_level=0).start(barrier=True)
            if energy_monitor is not None:
                energy_monitor.resume()

        # Miscellaneous post-training-step functions (e.g., FT heartbeats, GC).
        # Some of these only happen at specific iterations.
        maybe_synchronize_training_step(config.train.train_sync_interval, global_state.train_state.step)
        num_floating_point_operations_since_last_log_event = maybe_report_stragglers(
            config.logger.log_interval,
            bool(getattr(config.straggler, "log_straggler", False)),
            straggler_timer,
            global_state.train_state.step,
            num_floating_point_operations_since_last_log_event,
        )
        maybe_check_weight_hash_across_dp_replicas(
            model,
            config.train.check_weight_hash_across_dp_replicas_interval,
            global_state.train_state.step,
            should_toggle_forward_pre_hook,
        )
        handle_profiling_stop(
            config.profiling,
            global_state.train_state.step,
            torch.distributed.get_rank(),
            prof,
            nsys_nvtx_context,
        )
        maybe_run_manual_gc(
            config.train.manual_gc,
            config.train.manual_gc_interval,
            global_state.train_state.step,
        )

        # Checkpoint and decide whether to exit.
        should_exit = checkpoint_and_decide_exit(
            global_state,
            model,
            optimizer,
            scheduler,
            num_floating_point_operations_so_far,
            checkpointing_context,
            train_data_iterator,
        )
        if should_exit:
            break

    _delete_cuda_graphs(cuda_graph_helper)

    # Flush TensorBoard, WandB writers and one-logger.
    writer = global_state.tensorboard_logger
    if writer:
        writer.flush()

    # Close out pre-hooks if using distributed optimizer and overlapped param gather.
    if pre_hook_enabled:
        disable_forward_pre_hook(model)

    # This will finalize all unfinalized async request and terminate
    # a persistent async worker if persistent ckpt worker is enabled
    fault_tolerance.on_checkpointing_start(global_state)
    maybe_finalize_async_save(global_state=global_state, ckpt_cfg=config.checkpoint, blocking=True, terminate=True)
    fault_tolerance.on_checkpointing_end(global_state=global_state, is_async_finalization=True)

    # Shutdown NVRx straggler detection if enabled
    safe_shutdown_nvrx_straggler_manager(global_state.nvrx_straggler_manager)

    if energy_monitor is not None:
        energy_monitor.lap()
        total_energy = energy_monitor.get_total()
        print_rank_0(f"Total training energy (GPU): {total_energy / 1e6} MJ")
        energy_monitor.shutdown()

    # If any exit conditions (signal handler, duration, iterations) have been reached, exit.
    if should_exit:
        # Close NVIDIA DLFw Inspect if enabled
        tensor_inspect_end_if_enabled(config.tensor_inspect)
        maybe_finalize_async_save(global_state=global_state, ckpt_cfg=config.checkpoint, blocking=True, terminate=True)
        wandb_writer = global_state.wandb_logger
        if wandb_writer:
            wandb_writer.finish()
        fault_tolerance.shutdown(global_state)
        sys.exit(exit_code)

    # Close NVIDIA DLFw Inspect at clean finish
    tensor_inspect_end_if_enabled(config.tensor_inspect)

    if should_fire(callback_manager, "on_train_end"):
        callback_manager.fire(
            "on_train_end",
            CallbackContext(
                state=global_state,
                model=model,
                user_state=callback_manager.user_state,
                optimizer=optimizer,
                scheduler=scheduler,
            ),
        )


def train_step(
    forward_step_func: ForwardStepCallable,
    data_iterator: Optional[Union[RerunDataIterator, list[RerunDataIterator]]],
    model: list[MegatronModule],
    optimizer: MegatronOptimizer,
    scheduler: OptimizerParamScheduler,
    global_state: GlobalState,
    pg_collection: ProcessGroupCollection,
    forward_backward_func: Callable,
) -> tuple[dict[str, torch.Tensor], int, bool, bool, int, Optional[float], Optional[int]]:
    """Single training step.

    Args:
        forward_step_func: Function that performs a forward step (already wrapped if needed)
        data_iterator: Iterator over training data
        model: list of model chunks
        optimizer: Optimizer for model parameters
        scheduler: Learning rate scheduler
        global_state: Global training state
        pg_collection: Process group collection
        forward_backward_func: forward-backward function

    Returns:
        tuple containing:
        - loss_dict: Dictionary of reduced losses
        - skipped_iter: Whether the iteration was skipped (1) or not (0)
        - should_checkpoint: Whether a checkpoint should be saved
        - should_exit: Whether training should exit
        - exit_code: Exit code if should_exit is True
        - grad_norm: Gradient norm if available, None otherwise
        - num_zeros_in_grad: Number of zeros in gradient if available, None otherwise
        - max_attention_logit: Maximum attention logit if available, None otherwise
    """
    cfg: ConfigContainer = global_state.cfg
    timers = global_state.timers
    model_config = get_model_config(model[0])
    train_config = cfg.train
    optim_config = cfg.optimizer

    rerun_state_machine = get_rerun_state_machine()
    while rerun_state_machine.should_run_forward_backward(data_iterator):
        # Set grad to zero.
        for model_chunk in model:
            model_chunk.zero_grad_buffer()
        optimizer.zero_grad()

        _handle_mxfp8_param_buffer_copy(
            optimizer=optimizer,
            model=model,
            reuse_grad_buf_for_mxfp8_param_ag=cfg.optimizer.reuse_grad_buf_for_mxfp8_param_ag,
            overlap_param_gather=cfg.ddp.overlap_param_gather,
        )

        # Handle finetuning vs pretraining data consumption
        seq_length = model_config.seq_length  # Default for pretraining
        forward_backward_data_iterator = data_iterator  # Default for pretraining

        if cfg.dataset.dataloader_type == "batch":
            # Finetuning path to support variable-length sequences
            from megatron.bridge.data.finetuning import prepare_finetuning_batch
            from megatron.bridge.data.iterator_utils import make_data_iterator_list

            forward_backward_data_iterator, seq_length = prepare_finetuning_batch(
                data_iterator=data_iterator,
                num_microbatches=get_num_microbatches(),
                default_seq_length=model_config.seq_length,
                seq_key="tokens",
            )

            # Forward-backward pass.
            # Convert to list of iterators for virtual pipeline parallelism
            # With virtual PP, each model chunk needs independent access to the same microbatch
            forward_backward_data_iterator = make_data_iterator_list(
                model=model,
                data_iterator=forward_backward_data_iterator,
            )

        # [ModelOpt]: Pipeline-parallel Distillation stacks student and teacher tensors
        if not cfg.dist.use_decentralized_pg:
            adjust_tensor_shapes_fn = get_tensor_shapes_adjust_fn_for_distillation(
                model,
                seq_length=model_config.seq_length,
                micro_batch_size=train_config.micro_batch_size,
                decoder_seq_length=model_config.seq_length,
            )
        else:
            adjust_tensor_shapes_fn = None

        # Forward pass.
        p2p_communicator = P2PCommunicator(pp_group=pg_collection.pp, config=model_config)
        losses_reduced = forward_backward_func(
            forward_step_func=forward_step_func,
            data_iterator=forward_backward_data_iterator,
            model=model,
            num_microbatches=get_num_microbatches(),
            seq_length=seq_length,
            micro_batch_size=train_config.micro_batch_size,
            decoder_seq_length=seq_length,
            forward_only=False,
            adjust_tensor_shapes_fn=adjust_tensor_shapes_fn,
            p2p_communicator=p2p_communicator,
            pg_collection=pg_collection,
        )
    should_checkpoint, should_exit, exit_code = rerun_state_machine.should_checkpoint_and_exit()
    if should_exit:
        return {}, True, should_checkpoint, should_exit, exit_code, None, None, None

    # Empty unused memory.
    if train_config.empty_unused_memory_level >= 1:
        torch.cuda.empty_cache()

    # Update parameters.
    timers("optimizer", log_level=1).start(barrier=optim_config.barrier_with_L1_time)
    update_successful, grad_norm, num_zeros_in_grad = optimizer.step()

    # get max attention logit for logging and run clip_qk()
    # Part of MuonClip Optimizer step
    log_max_attention_logit = None
    if hasattr(cfg.model, "qk_clip") and cfg.model.qk_clip:
        log_max_attention_logit = clip_qk(model)

    timers("optimizer").stop()

    # when freezing sub-models we may have a mixture of successful and unsucessful ranks,
    # so we must gather across mp ranks
    update_successful = logical_and_across_model_parallel_group(update_successful, mp_group=pg_collection.mp)
    # grad_norm and num_zeros_in_grad will be None on ranks without trainable params,
    # so we must gather across mp ranks
    grad_norm = reduce_max_stat_across_model_parallel_group(grad_norm, mp_group=pg_collection.mp)
    if optim_config.log_num_zeros_in_grad:
        num_zeros_in_grad = reduce_max_stat_across_model_parallel_group(num_zeros_in_grad, mp_group=pg_collection.mp)

    # Update learning rate.
    if update_successful:
        increment = get_num_microbatches() * train_config.micro_batch_size * cfg.data_parallel_size
        scheduler.step(increment=increment)
        skipped_iter = 0
    else:
        skipped_iter = 1

    # Empty unused memory.
    if train_config.empty_unused_memory_level >= 2:
        torch.cuda.empty_cache()

    if is_pp_last_stage(pg_collection.pp):
        # Average loss across microbatches.
        loss_reduced = {}

        for key in losses_reduced[0].keys():
            val = [x[key].view(-1) for x in losses_reduced]
            if val[0].numel() == 2:
                # there is one dict per microbatch. in new reporting, we average
                # over the total number of tokens across the global batch.
                val = torch.vstack(val).sum(dim=0)
                dp_cp_group = pg_collection.dp_cp
                torch.distributed.all_reduce(val, group=dp_cp_group)
                loss_reduced[key] = val[0] / val[1]
            elif val[0].numel() == 1:
                # legacy behavior, we average over the number of microbatches
                val = torch.cat(val).mean()
                loss_reduced[key] = val
            else:
                raise ValueError(f"Invalid value shape: {val[0].shape} for key {key}")
        return (
            loss_reduced,
            skipped_iter,
            should_checkpoint,
            should_exit,
            exit_code,
            grad_norm,
            num_zeros_in_grad,
            log_max_attention_logit,
        )
    return (
        {},
        skipped_iter,
        should_checkpoint,
        should_exit,
        exit_code,
        grad_norm,
        num_zeros_in_grad,
        log_max_attention_logit,
    )


def maybe_synchronize_training_step(train_sync_interval: Optional[int], iteration: int) -> None:
    """Synchronizes CUDA streams when the configured interval is reached.

    Args:
        train_sync_interval: Number of iterations between synchronizations; ``None`` disables it.
        iteration: Zero-based training iteration counter.
    """

    if train_sync_interval and iteration % train_sync_interval == 0:
        torch.cuda.synchronize()


def maybe_report_stragglers(
    log_interval: int,
    log_straggler: bool,
    straggler_timer: Any,
    iteration: int,
    num_floating_point_operations_since_last_log_event: float,
) -> float:
    """Reports straggler metrics if logging is enabled.

    Args:
        log_interval: Iteration interval for logging.
        log_straggler: Whether straggler logging is enabled.
        straggler_timer: Timer utility used to record straggler metrics.
        iteration: Zero-based training iteration counter.
        num_floating_point_operations_since_last_log_event: FLOPs accumulated since the last
            logging event.

    Returns:
        float: Updated FLOP counter, reset to ``0.0`` when a report is emitted; otherwise the
        original value.
    """

    if log_straggler and log_interval:
        if iteration % log_interval == 0:
            straggler_timer.report(
                num_floating_point_operations_since_last_log_event,
                log_interval,
            )
            return 0.0
    return num_floating_point_operations_since_last_log_event


def maybe_check_weight_hash_across_dp_replicas(
    model: list[MegatronModule],
    check_weight_hash_across_dp_replicas_interval: Optional[int],
    iteration: int,
    should_toggle_forward_pre_hook: bool,
) -> None:
    """Verifies weight hashes across data-parallel replicas when requested.

    Args:
        model: List of model chunks to validate.
        check_weight_hash_across_dp_replicas_interval: Interval at which to verify; ``None`` to skip.
        iteration: Zero-based training iteration counter.
        should_toggle_forward_pre_hook: Whether the pre-hook must be disabled during the check.
    """

    interval = check_weight_hash_across_dp_replicas_interval
    if interval is None or iteration % interval != 0:
        return

    if should_toggle_forward_pre_hook:
        disable_forward_pre_hook(model)
    assert check_param_hashes_across_dp_replicas(model, cross_check=True), (
        "Parameter hashes not matching across DP replicas"
    )
    torch.distributed.barrier()
    print_rank_0(f">>> Weight hashes match after {iteration} iterations...")
    if should_toggle_forward_pre_hook:
        enable_forward_pre_hook(model)


def maybe_run_manual_gc(manual_gc_enabled: bool, manual_gc_interval: int, iteration: int) -> None:
    """Runs manual garbage collection according to the configured interval.

    Args:
        manual_gc_enabled: Whether manual garbage collection is enabled.
        manual_gc_interval: Number of iterations between collections; ``0`` disables periodic runs.
        iteration: Zero-based training iteration counter.
    """

    if manual_gc_enabled and manual_gc_interval != 0:
        if iteration % manual_gc_interval == 0:
            gc.collect()


def should_disable_forward_pre_hook(
    use_megatron_fsdp: bool, use_distributed_optimizer: bool, overlap_param_gather: bool
) -> bool:
    """Determine if forward pre-hooks should be disabled during checkpointing.

    Forward pre-hooks need to be disabled during checkpoint saving when using
    distributed optimizer with overlapped parameter gathering

    Args:
        use_megatron_fsdp: Whether Megatron FSDP is enabled.
        use_distributed_optimizer: Whether distributed optimizer is enabled.
        overlap_param_gather: Whether parameter gathering is overlapped.

    Returns:
        True if forward pre-hooks should be disabled, False otherwise.

    Note:
        This is needed to prevent autograd issues during checkpoint saving
        when using distributed optimizer with parameter gathering overlap.
    """
    return not use_megatron_fsdp and use_distributed_optimizer and overlap_param_gather


def enable_forward_pre_hook(model: list[DDP]) -> None:
    """Enable forward pre-hook for all model chunks.

    Args:
        model: list of model chunks wrapped in DDP
    """
    for model_chunk in model:
        assert isinstance(model_chunk, DDP)
        model_chunk.enable_forward_pre_hook()


def disable_forward_pre_hook(model: list[DDP], param_sync: bool = True) -> None:
    """Disable forward pre-hook for all model chunks.

    Args:
        model: list of model chunks wrapped in DDP
        param_sync: Whether to synchronize parameters across model chunks
    """
    for model_chunk in model:
        assert isinstance(model_chunk, DDP)
        model_chunk.disable_forward_pre_hook(param_sync=param_sync)


def force_param_sync(model: list[DDP]) -> None:
    """Force parameter synchronization for all model chunks.

    Args:
        model: list of model chunks wrapped in DDP.
    """
    for model_chunk in model:
        assert isinstance(model_chunk, DDP)
        model_chunk.start_param_sync(force_sync=True)


def get_start_time_from_progress_log(cfg: ConfigContainer) -> tuple[datetime, float]:
    """
    Gets start time of earliest job with same world size. Also returns the number
    of floating-point operations completed in last saved checkpoint.
    """
    assert cfg.checkpoint.save is not None
    progress_log_filename = os.path.join(cfg.checkpoint.save, "progress.txt")

    # start_time is time when job with same world size started.
    # start_num_floating_point_operations is the number of floating-point operations
    # completed when this job started.
    # latest_num_floating_point_operations is the number of floating-point operations
    # completed in most recent saved checkpoint.
    start_time = None
    start_num_floating_point_operations = None
    latest_num_floating_point_operations = 0

    def _get_field(string, type):
        return type(string.split(": ")[1])

    with open(progress_log_filename, "r") as f:
        for line in f:
            line = line.strip()
            line_tokens = line.split("\t")
            world_size_in_line = _get_field(line_tokens[2], int)
            if line_tokens[3] == "Saved checkpoint":
                latest_num_floating_point_operations = _get_field(line_tokens[7], float)
            if world_size_in_line != get_world_size_safe():
                # Re-start search if we see a different world size.
                start_time = None
                start_num_floating_point_operations = None
                continue
            if line_tokens[3] == "Starting job":
                if start_time is None:
                    start_time = line_tokens[0]
                    start_num_floating_point_operations = latest_num_floating_point_operations
    assert start_time is not None and start_num_floating_point_operations is not None, (
        "Should have seen at least one 'Starting job' entry with same world_size"
    )
    return datetime.strptime(start_time, "%Y-%m-%d %H:%M:%S"), start_num_floating_point_operations


def compute_throughputs_and_append_to_progress_log(
    state: GlobalState, num_floating_point_operations_so_far: float
) -> None:
    """Computes job and cumulative throughputs and appends to progress log.

    Calculates Model TFLOP/s/GPU based on floating-point operations and elapsed time.
    Appends the computed throughputs, total FLOPs, and processed tokens to the
    progress log file.

    Args:
        state: The GlobalState object.
        num_floating_point_operations_so_far: Total floating-point operations completed.
    """
    if state.cfg.checkpoint.save is None:
        return

    # Compute job throughput.
    # num_floating_point_operations_so_far keeps track of floating-point operations
    # completed at the start of job.
    job_throughput = (num_floating_point_operations_so_far - state.train_state.floating_point_operations_so_far) / (
        (time.time() - state.start_time) * 10**12 * get_world_size_safe()
    )

    # Compute cumulative throughput since jobs of this world size were launched.
    # `get_start_time_from_progress_log` returns start time and number of floating-point
    # operations of first job of this world size.
    start_time, start_num_floating_point_operations = get_start_time_from_progress_log(state.cfg)
    elapsed_time = (datetime.now() - start_time).total_seconds()
    cumulative_throughput = (num_floating_point_operations_so_far - start_num_floating_point_operations) / (
        elapsed_time * 10**12 * get_world_size_safe()
    )

    tokens_so_far = state.train_state.consumed_train_samples * state.cfg.model.seq_length
    saved_ckpt_prefix = "Saving async checkpoint" if state.cfg.checkpoint.async_save else "Saved checkpoint"
    append_to_progress_log(
        state.cfg.checkpoint.save,
        f"{saved_ckpt_prefix}\tIteration: {state.train_state.step}\t"
        f"Job throughput: {job_throughput:.1f} MODEL_TFLOP/s/GPU\t"
        f"Cumulative throughput: {cumulative_throughput:.1f} MODEL_TFLOP/s/GPU\t"
        f"Floating-point operations: {num_floating_point_operations_so_far:.2e}\t"
        f"Tokens (in billions): {tokens_so_far / 10**9:.2f}",
    )


def save_checkpoint_and_time(
    state: GlobalState,
    model: list[MegatronModule],
    optimizer: MegatronOptimizer,
    opt_param_scheduler: OptimizerParamScheduler,
    num_floating_point_operations_so_far: float,
    checkpointing_context: dict[str, Any],
    non_persistent_ckpt: bool = False,
    train_data_iterator: Optional[Union[RerunDataIterator, list[RerunDataIterator]]] = None,
) -> None:
    """Saves a checkpoint and logs the timing.

    Wraps the `save_checkpoint` function with timers and forces parameter
    synchronization when using distributed optimizer with overlapped parameter
    gather to ensure checkpoint correctness.

    Args:
        state: The global state object.
        model: list of model chunks (MegatronModule instances).
        optimizer: The optimizer instance.
        opt_param_scheduler: The optimizer parameter scheduler instance.
        num_floating_point_operations_so_far: Cumulative Model TFLOPs up to this point.
        checkpointing_context: Dictionary holding checkpointing-related state.
        non_persistent_ckpt: Flag indicating if this is a non-persistent
                             (local) checkpoint. Defaults to False.
        train_data_iterator: Optional training data iterator to save its state.
    """
    timers = state.timers
    energy_monitor = state.energy_monitor

    # Stop timer to get accurate train interval time and exclude checkpointing duration
    timers("interval-time").stop()

    # Pause energy monitor
    if energy_monitor is not None:
        energy_monitor.pause()

    # Extra barrier is added to make sure all ranks report the max time.
    timer_key = "save-checkpoint-non-persistent" if non_persistent_ckpt else "save-checkpoint"
    timers(timer_key, log_level=0).start(barrier=True)

    should_force_param_sync = should_disable_forward_pre_hook(
        state.cfg.ddp.use_megatron_fsdp,
        state.cfg.optimizer.use_distributed_optimizer,
        state.cfg.ddp.overlap_param_gather,
    )
    if should_force_param_sync:
        force_param_sync(model)
    save_checkpoint(
        state,
        model,
        optimizer,
        opt_param_scheduler,
        num_floating_point_operations_so_far,
        checkpointing_context=checkpointing_context,
        non_persistent_ckpt=non_persistent_ckpt,
        train_data_iterator=train_data_iterator,
    )
    if state.cfg.model.fp8 is not None:
        # Run garbage collection after checkpoint saving to free memory from
        # dequantized bf16 tensors that were temporarily created during fp8
        # model checkpoint saving.
        gc.collect()
    timers(timer_key).stop(barrier=True)
    timers.log([timer_key])

    if state.cfg.logger.log_progress and not non_persistent_ckpt:
        compute_throughputs_and_append_to_progress_log(state, num_floating_point_operations_so_far)

    # Recover timing
    if energy_monitor is not None:
        energy_monitor.resume()
    timers("interval-time", log_level=0).start(barrier=True)


def checkpoint_and_decide_exit(
    state: GlobalState,
    model: list[MegatronModule],
    optimizer: MegatronOptimizer,
    opt_param_scheduler: OptimizerParamScheduler,
    num_floating_point_operations_so_far: float,
    checkpointing_context: dict[str, Any],
    train_data_iterator: Optional[Union[RerunDataIterator, list[RerunDataIterator]]],
) -> bool:
    """Handles checkpointing decisions and determines if training should exit.

    Checks various conditions for saving a checkpoint (signal received, interval,
    duration) and determines if the training loop should terminate based on exit
    conditions (signal, duration, iteration interval).

    Args:
        state: The global state object.
        model: list of model chunks (MegatronModule instances).
        optimizer: The optimizer instance.
        opt_param_scheduler: The optimizer parameter scheduler instance.
        num_floating_point_operations_so_far: Cumulative TFLOPs up to this point.
        checkpointing_context: Dictionary holding checkpointing-related state.
        train_data_iterator: Optional training data iterator to save its state.

    Returns:
        True if the training loop should exit, False otherwise.
    """
    saved_checkpoint = False

    # Exit based on signal handler.
    if state.cfg.train.exit_signal_handler:
        signal_handler = state.signal_handler
        if any(signal_handler.signals_received()):
            if state.cfg.checkpoint.save:
                save_checkpoint_and_time(
                    state,
                    model,
                    optimizer,
                    opt_param_scheduler,
                    num_floating_point_operations_so_far,
                    checkpointing_context,
                    train_data_iterator=train_data_iterator,
                )
            barrier_and_log("exiting program after receiving SIGTERM.")

            return True

    # Regular save (persistent and non-persistent).
    if (
        state.cfg.checkpoint.save
        and state.cfg.checkpoint.save_interval
        and state.train_state.step % state.cfg.checkpoint.save_interval == 0
    ):
        save_checkpoint_and_time(
            state,
            model,
            optimizer,
            opt_param_scheduler,
            num_floating_point_operations_so_far,
            checkpointing_context,
            train_data_iterator=train_data_iterator,
        )
        saved_checkpoint = True

    elif (
        state.cfg.checkpoint.save
        and state.cfg.checkpoint.non_persistent_save_interval
        and state.train_state.step % state.cfg.checkpoint.non_persistent_save_interval == 0
    ):
        save_checkpoint_and_time(
            state,
            model,
            optimizer,
            opt_param_scheduler,
            num_floating_point_operations_so_far,
            checkpointing_context,
            non_persistent_ckpt=True,
            train_data_iterator=train_data_iterator,
        )
        saved_checkpoint = True

    # Exit based on duration.
    if state.cfg.train.exit_duration_in_mins:
        train_time = (time.time() - state.start_time) / 60.0
        done_cuda = torch.tensor([train_time > state.cfg.train.exit_duration_in_mins], dtype=torch.int, device="cuda")
        torch.distributed.all_reduce(done_cuda, op=torch.distributed.ReduceOp.MAX)
        done = done_cuda.item()
        if done:
            if state.cfg.checkpoint.save and not saved_checkpoint:
                save_checkpoint_and_time(
                    state,
                    model,
                    optimizer,
                    opt_param_scheduler,
                    num_floating_point_operations_so_far,
                    checkpointing_context,
                    train_data_iterator=train_data_iterator,
                )
            barrier_and_log(f"exiting program after {train_time} minutes")

            return True

    # Exit based on iterations.
    if state.cfg.train.exit_interval and state.train_state.step % state.cfg.train.exit_interval == 0:
        if state.cfg.checkpoint.save and not saved_checkpoint:
            save_checkpoint_and_time(
                state,
                model,
                optimizer,
                opt_param_scheduler,
                num_floating_point_operations_so_far,
                checkpointing_context,
                train_data_iterator=train_data_iterator,
            )
        barrier_and_log(f"exiting program at iteration {state.train_state.step}")

        return True

    # Exit based on NVRx straggler detection
    if check_nvrx_straggler_detection(state.nvrx_straggler_manager):
        if state.cfg.checkpoint.save is not None and not saved_checkpoint:
            save_checkpoint_and_time(
                state,
                model,
                optimizer,
                opt_param_scheduler,
                num_floating_point_operations_so_far,
                checkpointing_context,
                train_data_iterator=train_data_iterator,
            )
        barrier_and_log("Exiting program due to straggler detection.")
        return True

    return False


def _finish_train(global_state: GlobalState):
    ckpt_cfg = global_state.cfg.checkpoint

    # Shutdown NVRx straggler detection if enabled
    safe_shutdown_nvrx_straggler_manager(global_state.nvrx_straggler_manager)

    fault_tolerance.on_checkpointing_start(global_state)
    maybe_finalize_async_save(global_state=global_state, blocking=True, terminate=True, ckpt_cfg=ckpt_cfg)
    fault_tolerance.on_checkpointing_end(global_state=global_state, is_async_finalization=True)
    fault_tolerance.shutdown(global_state)

    if global_state.wandb_logger:
        global_state.wandb_logger.finish()

    destroy_global_state()


def _should_skip_and_handle_iteration(
    global_state: GlobalState,
    train_data_iterator: Optional[Union[RerunDataIterator, list[RerunDataIterator]]],
    pg_collection: ProcessGroupCollection,
) -> bool:
    """Check if the current iteration should be skipped and handle it if so.

    This function checks if the current training step is in the iterations_to_skip list,
    and if so, performs a dummy training step to consume data and update counters.

    Args:
        global_state: Global state containing training state and configuration
        train_data_iterator: Iterator over training data

    Returns:
        bool: True if the iteration was skipped, False otherwise
    """
    cfg = global_state.cfg
    if global_state.train_state.step not in cfg.train.iterations_to_skip:
        return False

    # Perform dummy train step to fast forward train_data_iterator
    _dummy_train_step(global_state, train_data_iterator, pg_collection)

    # Update step and sample counters
    global_state.train_state.step += 1
    dp_size = pg_collection.dp.size()
    batch_size = dp_size * cfg.train.micro_batch_size * get_num_microbatches()
    global_state.train_state.consumed_train_samples += batch_size
    global_state.train_state.skipped_train_samples += batch_size

    return True


def _dummy_train_step(
    global_state: GlobalState,
    train_data_iterator: Optional[Union[RerunDataIterator, list[RerunDataIterator]]],
    pg_collection: ProcessGroupCollection,
) -> None:
    """Single dummy training step to fast forward train_data_iterator.

    This function consumes data from the iterator without performing any actual computation,
    effectively skipping the iteration while maintaining data iterator consistency.

    Advance the data iterator on first and last PP stages when data_iterator is not None.

    Args:
        global_state: Global state containing configuration
        train_data_iterator: Iterator over training data
    """
    cfg = global_state.cfg
    num_microbatches = get_num_microbatches()
    rerun_state_machine = get_rerun_state_machine()

    while rerun_state_machine.should_run_forward_backward(train_data_iterator):
        pp_group = pg_collection.pp
        if is_pp_first_stage(pp_group) or is_pp_last_stage(pp_group):
            if train_data_iterator is not None:
                if cfg.dataset.dataloader_type == "batch":
                    # Finetuning: Consume global batch once
                    _ = next(train_data_iterator)
                else:
                    # Pretrain: Consume microbatches one at a time
                    for _ in range(num_microbatches):
                        _ = next(train_data_iterator)


def _handle_mxfp8_param_buffer_copy(
    optimizer: MegatronOptimizer,
    model: list[MegatronModule],
    reuse_grad_buf_for_mxfp8_param_ag: bool,
    overlap_param_gather: bool,
) -> None:
    """Copy main params to param buffer for mxfp8 with grad buffer reuse.

    For mxfp8_param with reuse_grad_buf_for_mxfp8_param_ag and dp_ag_overlap,
    we need to call _copy_main_params_to_param_buffer() after the grad buffer
    is zeroed because param and grad buffer are shared.

    However, we should skip this on the first iteration when forward_pre_hook is disabled,
    because:
    1. The first iteration's params are already in param.data (from init or checkpoint).
    2. Without forward_pre_hook, finish_param_sync() won't be called to zero the grad buffer,
       so the main grads will be polluted by the main params.

    Args:
        optimizer: The MegatronOptimizer instance
        model: List of model chunks (MegatronModule instances)
        reuse_grad_buf_for_mxfp8_param_ag: Config flag for grad buffer reuse
        overlap_param_gather: Config flag for overlapping param gathering
    """
    if reuse_grad_buf_for_mxfp8_param_ag and overlap_param_gather:
        # Check if forward_pre_hook is enabled by checking if hooks are registered.
        forward_pre_hook_enabled = len(model[0].remove_forward_pre_hook_handles) > 0
        if forward_pre_hook_enabled:
            for optim_instance in optimizer.chained_optimizers:
                if isinstance(optim_instance, DistributedOptimizer):
                    optim_instance._copy_main_params_to_param_buffer()


def _delete_cuda_graphs(cuda_graph_helper: TECudaGraphHelper):
    """
    Delete the CUDA graph object as they hold a reference to the some of the nccl buffers, thus blocking the
    process-destory (torch.dist.destroy_process_group()) at the end of the training loop.

    TODO: Move this method to MCore.

    Args:
        cuda_graph_helper: The TECudaGraphHelper object.

    """

    print_rank_0("Deleting CUDA graphs")

    # Explicitly delete the training CUDA graph because of
    # https://github.com/pytorch/pytorch/issues/115388#issuecomment-3009880966
    if "training" in FullCudaGraphWrapper.cuda_graph:
        del FullCudaGraphWrapper.cuda_graph["training"]

    # Cleanup CUDA graphs object for partial Cuda-graphs (implemented in TransformerEngine)
    if cuda_graph_helper is not None:
        for layers in cuda_graph_helper.callables_per_chunk:
            for layer in layers:
                for cuda_graph in layer.cuda_graphs:
                    del cuda_graph
                del layer.cuda_graphs

    # Run GC to collect the freshed object
    gc.collect()


def _maybe_register_fsdp_buffers(
    config: ConfigContainer,
    model: list[MegatronModule],
) -> None:
    """Manually register FSDP communication buffers if enabled."""
    # If fsdp_manual_registration is enabled, manually register FSDP communication buffers after one training step.
    if (
        config.ddp.use_megatron_fsdp
        and hasattr(config.ddp, "fsdp_manual_registration")
        and config.ddp.fsdp_manual_registration
    ):
        print_rank_0("[Megatron-FSDP] Registering FSDP communication buffers manually")
        for model_chunk in model:
            if isinstance(model_chunk, megatron_FSDP) and getattr(
                model_chunk.ddp_config, "fsdp_manual_registration", False
            ):
                fsdp_param_and_grad_buffer = getattr(model_chunk, "param_and_grad_buffer", None)
                if fsdp_param_and_grad_buffer is not None:
                    fsdp_param_and_grad_buffer.manual_buffer_registration()
        print_rank_0("[Megatron-FSDP] Buffer registered")



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from typing import Optional, Union

import torch

from megatron.bridge import AutoBridge
from megatron.bridge.data.vlm_datasets import (
    HFDatasetConversationProvider,
)
from megatron.bridge.data.vlm_datasets.mock_provider import MockVLMConversationProvider
from megatron.bridge.peft.lora import VLMLoRA
from megatron.bridge.recipes.utils.optimizer_utils import distributed_fused_adam_with_cosine_annealing
from megatron.bridge.recipes.utils.tokenizer_utils import DEFAULT_NULL_TOKENIZER_VOCAB_SIZE
from megatron.bridge.training.comm_overlap import CommOverlapConfig
from megatron.bridge.training.config import (
    CheckpointConfig,
    ConfigContainer,
    DistributedDataParallelConfig,
    LoggerConfig,
    RNGConfig,
    TokenizerConfig,
    TrainingConfig,
)
from megatron.bridge.training.mixed_precision import MixedPrecisionConfig


def nemotron_nano_v2_vl_12b_pretrain_config(
    dir: Optional[str] = None,
    name: str = "nemotron_nano_v2_vl_pretrain",
    hf_model_path: str = "nvidia/NVIDIA-Nemotron-Nano-12B-v2-VL-BF16",
    # Dataset configuration
    dataset_type: Optional[str] = None,
    mock: bool = False,
    dataset_maker_name: str = "make_cord_v2_dataset",
    # Model configuration
    tensor_parallelism: int = 4,
    pipeline_parallelism: int = 1,
    pipeline_parallelism_dtype: Optional[torch.dtype] = None,
    virtual_pipeline_parallelism: Optional[int] = None,
    context_parallelism: int = 1,
    sequence_parallelism: bool = False,
    # Training hyperparameters
    train_iters: int = 300000,
    global_batch_size: int = 32,
    micro_batch_size: int = 2,
    seq_length: int = 4096,
    lr: float = 3e-4,
    min_lr: float = 3e-5,
    lr_warmup_iters: int = 500,
    lr_decay_iters: Optional[int] = None,
    # Precision and comm overlap
    precision_config: Optional[Union[MixedPrecisionConfig, str]] = "bf16_mixed",
    comm_overlap_config: Optional[CommOverlapConfig] = None,
    # Checkpointing
    save_interval: Optional[int] = 200,
) -> ConfigContainer:
    """
    Create a pre-training configuration for Nemotron Nano V2 VL.

    Note: Current dataset pipeline is text-centric. To train multimodal tokens,
    your preprocessed data should include placeholder tokens (e.g., <image>) as needed.
    """
    base_output_dir = dir if dir is not None else os.path.join(os.getcwd(), "nemo_experiments")
    run_output_dir = os.path.join(base_output_dir, name)
    checkpoint_dir = os.path.join(run_output_dir, "checkpoints")
    tensorboard_dir = os.path.join(run_output_dir, "tb_logs")

    # Build provider via AutoBridge and set parallel/seq params here
    bridge = AutoBridge.from_hf_pretrained(hf_model_path, trust_remote_code=True)
    model_cfg = bridge.to_megatron_provider(load_weights=False)
    model_cfg.tensor_model_parallel_size = tensor_parallelism
    model_cfg.pipeline_model_parallel_size = pipeline_parallelism
    model_cfg.pipeline_dtype = pipeline_parallelism_dtype
    model_cfg.virtual_pipeline_model_parallel_size = virtual_pipeline_parallelism
    model_cfg.context_parallel_size = context_parallelism
    model_cfg.sequence_parallel = sequence_parallelism
    model_cfg.seq_length = seq_length

    opt_config, scheduler = distributed_fused_adam_with_cosine_annealing(
        lr_warmup_iters=lr_warmup_iters,
        lr_decay_iters=lr_decay_iters,
        max_lr=lr,
        min_lr=min_lr,
    )

    # Dataset provider selection
    _dataset_choice = (dataset_type or ("mock" if mock else "hf")).lower()

    if _dataset_choice == "mock":
        dataset_cfg = MockVLMConversationProvider(
            seq_length=seq_length,
            hf_processor_path=hf_model_path,
            dataloader_type="single",
        )
    elif _dataset_choice == "hf":
        dataset_cfg = HFDatasetConversationProvider(
            seq_length=seq_length,
            hf_processor_path=hf_model_path,
            maker_name=dataset_maker_name,
            # Dataloader config parameters
            num_workers=2,
            dataloader_type="single",
            data_sharding=True,
            pin_memory=True,
            persistent_workers=False,
        )
    else:
        raise ValueError(f"Unknown dataset_type '{_dataset_choice}'. Expected one of: 'mock', 'hf', 'preloaded'.")

    # Config Container
    cfg = ConfigContainer(
        model=model_cfg,
        train=TrainingConfig(
            train_iters=train_iters,
            eval_interval=500,
            eval_iters=32,
            global_batch_size=global_batch_size,
            micro_batch_size=micro_batch_size,
            manual_gc=True,
            manual_gc_interval=100,
            manual_gc_eval=100,
        ),
        optimizer=opt_config,
        scheduler=scheduler,
        ddp=DistributedDataParallelConfig(
            check_for_nan_in_grad=True,
            grad_reduce_in_fp32=True,
            overlap_grad_reduce=False,
            overlap_param_gather=False,
            average_in_collective=False,
            data_parallel_sharding_strategy="optim_grads_params",
            use_distributed_optimizer=True,
            # use_megatron_fsdp=use_megatron_fsdp,  # need use_distributed_optimizer=True
        ),
        dataset=dataset_cfg,
        logger=LoggerConfig(
            log_interval=10,
            tensorboard_dir=tensorboard_dir,
            log_timers_to_tensorboard=True,
        ),
        tokenizer=TokenizerConfig(tokenizer_type="NullTokenizer", vocab_size=DEFAULT_NULL_TOKENIZER_VOCAB_SIZE),
        checkpoint=CheckpointConfig(
            save_interval=save_interval,
            save=checkpoint_dir,
            load=checkpoint_dir,
            ckpt_format="torch_dist",
            fully_parallel_save=True,
        ),
        rng=RNGConfig(seed=1234),
        comm_overlap=comm_overlap_config,
        mixed_precision=precision_config,
    )

    return cfg


def nemotron_nano_v2_vl_12b_finetune_config(
    *,
    pretrained_checkpoint: str = "",
    lora_on_language_model: bool = False,
    lora_on_vision_model: bool = False,
    save_checkpoint_dir: Optional[str] = None,
    **pretrain_kwargs,
) -> ConfigContainer:
    """Create a finetuning configuration for Nemotron Nano V2 VL.

    This helper wraps :func:`nemotron_nano_v2_vl_12b_pretrain_config`, forwarding all keyword arguments to it
    while additionally wiring up the :class:`CheckpointConfig` for finetuning from a
    given *``pretrained_checkpoint``*.

    Parameters:
    pretrained_checkpoint: str
        Path to a Megatron-Bridge checkpoint (or a directory produced by
        ``convert_ckpt_hf_to_megatron``) that will be loaded before training.
    save_checkpoint_dir: str | None, default ``run_output_dir / "checkpoints"``
        Directory where new checkpoints will be saved / resumed from.  If not
        provided, we reuse the default path chosen by *nemotron_nano_v2_vl_12b_pretrain_config*.
    lora_on_language_model: bool = True
        Whether to apply PEFT to the language model.
    lora_on_vision_model: bool = True
        Whether to apply PEFT to the vision model.
    **pretrain_kwargs: Any
        Additional keyword arguments are forwarded verbatim to
        :func:`nemotron_nano_v2_vl_12b_pretrain_config` to customise the base recipe (e.g. batch size,
        learning rate, parallelism).
    """

    cfg = nemotron_nano_v2_vl_12b_pretrain_config(**pretrain_kwargs)

    # Override Train hyper-parameters suitable for finetuning if the caller did
    # not explicitly pass them via **pretrain_kwargs.
    if pretrain_kwargs.get("train_iters") is None:
        cfg.train.train_iters = 10_000
    if pretrain_kwargs.get("lr") is None and hasattr(cfg.optimizer, "lr"):
        cfg.optimizer.lr = 1e-5  # type: ignore[attr-defined]
    if pretrain_kwargs.get("min_lr") is None and hasattr(cfg.optimizer, "min_lr"):
        cfg.optimizer.min_lr = 1e-6  # type: ignore[attr-defined]

    # Update CheckpointConfig for finetuning.
    ckpt_dir = save_checkpoint_dir or cfg.checkpoint.save or cfg.checkpoint.load  # type: ignore[attr-defined]
    cfg.checkpoint = CheckpointConfig(
        pretrained_checkpoint=pretrained_checkpoint,
        save=ckpt_dir,
        load=ckpt_dir,
        ckpt_format=cfg.checkpoint.ckpt_format,  # preserve existing choice
        fully_parallel_save=cfg.checkpoint.fully_parallel_save,
        save_interval=cfg.checkpoint.save_interval,
    )
    if lora_on_language_model:
        if lora_on_vision_model:
            cfg.peft = VLMLoRA(
                target_modules=["linear_qkv", "linear_proj", "linear_fc1", "linear_fc2"],
                dim=16,
                alpha=32,
            )
        else:
            cfg.peft = VLMLoRA(
                target_modules=[
                    "*language_model*.linear_qkv",
                    "*language_model*.linear_proj",
                    "*language_model*.linear_fc1",
                    "*language_model*.linear_fc2",
                ],
                dim=16,
                alpha=32,
                freeze_vision_model=False,
                freeze_vision_projection=False,
            )

        cfg.optimizer.lr = 5e-5
        cfg.optimizer.min_lr = 5e-6
        cfg.model.tensor_model_parallel_size = 2

    return cfg



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This file is taken from https://github.com/NVIDIA/NeMo-Curator, which is adapted from cuML's safe_imports module:
# https://github.com/rapidsai/cuml/blob/e93166ea0dddfa8ef2f68c6335012af4420bc8ac/python/cuml/internals/safe_imports.py


import importlib
import logging
import traceback
from contextlib import contextmanager
from typing import Tuple

import torch
from packaging.version import Version as PkgVersion


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())

GPU_INSTALL_STRING = (
    """Install GPU packages via `pip install --extra-index-url """
    """https://pypi.nvidia.com nemo-curator[cuda12x]`
or use `pip install --extra-index-url https://pypi.nvidia.com ".[cuda12x]"` if installing from source"""
)
MISSING_NEMO_EXPORT_DEPLOY_MSG = (
    "nemo-export-deploy is not available. Please install it with `pip install nemo-export-deploy`."
)
MISSING_NVRX_MSG = (
    "nvidia-resiliency-ext is not available. Please install it with `pip install nvidia-resiliency-ext`."
)
MISSING_NEMO_RUN_MSG = "nemo-run is not available. Please install it with `pip install nemo-run`."


class UnavailableError(Exception):
    """Error thrown if a symbol is unavailable due to an issue importing it"""


@contextmanager
def null_decorator(*args, **kwargs):
    """null_decorator"""
    if len(kwargs) == 0 and len(args) == 1 and callable(args[0]):
        return args[0]
    else:

        def inner(func):
            return func

        return inner


class UnavailableMeta(type):
    """A metaclass for generating placeholder objects for unavailable symbols

    This metaclass allows errors to be deferred from import time to the time
    that a symbol is actually used in order to streamline the usage of optional
    dependencies. This is particularly useful for attempted imports of GPU-only
    modules which will only be invoked if GPU-only functionality is
    specifically used.

    If an attempt to import a symbol fails, this metaclass is used to generate
    a class which stands in for that symbol. Any attempt to call the symbol
    (instantiate the class) or access its attributes will throw an
    UnavailableError exception. Furthermore, this class can be used in
    e.g. isinstance checks, since it will (correctly) fail to match any
    instance it is compared against.

    In addition to calls and attribute access, a number of dunder methods are
    implemented so that other common usages of imported symbols (e.g.
    arithmetic) throw an UnavailableError, but this is not guaranteed for
    all possible uses. In such cases, other exception types (typically
    TypeErrors) will be thrown instead.
    """

    def __new__(meta, name, bases, dct):
        if dct.get("_msg", None) is None:
            dct["_msg"] = f"{name} could not be imported"
        name = f"MISSING{name}"
        return super(UnavailableMeta, meta).__new__(meta, name, bases, dct)

    def __call__(cls, *args, **kwargs):
        raise UnavailableError(cls._msg)

    def __getattr__(cls, name):
        # Special handling for unittest.mock which tries to access __func_
        # and other attributes during its operations
        if name in ("__func__", "__wrapped__", "__name__", "__qualname__"):
            raise AttributeError(f"'{cls.__name__}' has no attribute '{name}'")
        raise UnavailableError(cls._msg)

    def __eq__(cls, other):
        raise UnavailableError(cls._msg)

    def __lt__(cls, other):
        raise UnavailableError(cls._msg)

    def __gt__(cls, other):
        raise UnavailableError(cls._msg)

    def __le__(cls, other):
        raise UnavailableError(cls._msg)

    def __ge__(cls, other):
        raise UnavailableError(cls._msg)

    def __ne__(cls, other):
        raise UnavailableError(cls._msg)

    def __abs__(cls):
        raise UnavailableError(cls._msg)

    def __add__(cls, other):
        raise UnavailableError(cls._msg)

    def __radd__(cls, other):
        raise UnavailableError(cls._msg)

    def __iadd__(cls, other):
        raise UnavailableError(cls._msg)

    def __floordiv__(cls, other):
        raise UnavailableError(cls._msg)

    def __rfloordiv__(cls, other):
        raise UnavailableError(cls._msg)

    def __ifloordiv__(cls, other):
        raise UnavailableError(cls._msg)

    def __lshift__(cls, other):
        raise UnavailableError(cls._msg)

    def __rlshift__(cls, other):
        raise UnavailableError(cls._msg)

    def __mul__(cls, other):
        raise UnavailableError(cls._msg)

    def __rmul__(cls, other):
        raise UnavailableError(cls._msg)

    def __imul__(cls, other):
        raise UnavailableError(cls._msg)

    def __ilshift__(cls, other):
        raise UnavailableError(cls._msg)

    def __pow__(cls, other):
        raise UnavailableError(cls._msg)

    def __rpow__(cls, other):
        raise UnavailableError(cls._msg)

    def __ipow__(cls, other):
        raise UnavailableError(cls._msg)

    def __rshift__(cls, other):
        raise UnavailableError(cls._msg)

    def __rrshift__(cls, other):
        raise UnavailableError(cls._msg)

    def __irshift__(cls, other):
        raise UnavailableError(cls._msg)

    def __sub__(cls, other):
        raise UnavailableError(cls._msg)

    def __rsub__(cls, other):
        raise UnavailableError(cls._msg)

    def __isub__(cls, other):
        raise UnavailableError(cls._msg)

    def __truediv__(cls, other):
        raise UnavailableError(cls._msg)

    def __rtruediv__(cls, other):
        raise UnavailableError(cls._msg)

    def __itruediv__(cls, other):
        raise UnavailableError(cls._msg)

    def __divmod__(cls, other):
        raise UnavailableError(cls._msg)

    def __rdivmod__(cls, other):
        raise UnavailableError(cls._msg)

    def __neg__(cls):
        raise UnavailableError(cls._msg)

    def __invert__(cls):
        raise UnavailableError(cls._msg)

    def __hash__(cls):
        raise UnavailableError(cls._msg)

    def __index__(cls):
        raise UnavailableError(cls._msg)

    def __iter__(cls):
        raise UnavailableError(cls._msg)

    def __delitem__(cls, name):
        raise UnavailableError(cls._msg)

    def __setitem__(cls, name, value):
        raise UnavailableError(cls._msg)

    def __enter__(cls, *args, **kwargs):
        raise UnavailableError(cls._msg)

    def __get__(cls, *args, **kwargs):
        raise UnavailableError(cls._msg)

    def __delete__(cls, *args, **kwargs):
        raise UnavailableError(cls._msg)

    def __len__(cls):
        raise UnavailableError(cls._msg)


def is_unavailable(obj):
    """Helper to check if given symbol is actually a placeholder"""
    return type(obj) is UnavailableMeta


class UnavailableNullContext:
    """A placeholder class for unavailable context managers

    This context manager will return a value which will throw an
    UnavailableError if used in any way, but the context manager itself can be
    safely invoked.
    """

    def __init__(self, *args, **kwargs):
        pass

    def __enter__(self):
        return UnavailableMeta(
            "MissingContextValue",
            (),
            {"_msg": "Attempted to make use of placeholder context return value."},
        )

    def __exit__(self, *args, **kwargs):
        pass


def safe_import(module, *, msg=None, alt=None) -> Tuple[object, bool]:
    """A function used to import modules that may not be available.

    This function will attempt to import a module with the given name, but it
    will not throw an ImportError if the module is not found. Instead, it will
    return a placeholder object which will raise an exception only if used.

    Args:
        module (str): The name of the module to import.
        msg (str, optional): An error message to be displayed if this module is used
            after a failed import. Defaults to None.
        alt (object, optional): A module to be used in place of the given module if it
            fails to import. Defaults to None.

    Returns:
        tuple: A tuple containing two elements. The first element is the imported module,
        the given alternate, or a class derived from UnavailableMeta. The second element
        is a boolean indicating whether the intended import was successful.
    """
    try:
        return importlib.import_module(module), True
    except ImportError:
        exception_text = traceback.format_exc()
        logger.debug(f"Import of {module} failed with: {exception_text}")
    except Exception:
        exception_text = traceback.format_exc()
        raise
    if msg is None:
        msg = f"{module} could not be imported"
    if alt is None:
        return UnavailableMeta(module.rsplit(".")[-1], (), {"_msg": msg}), False
    else:
        return alt, False


def safe_import_from(module, symbol, *, msg=None, alt=None, fallback_module=None) -> Tuple[object, bool]:
    """A function used to import symbols from modules that may not be available.

    This function will attempt to import a symbol with the given name from
    the given module, but it will not throw an ImportError if the symbol is not
    found. Instead, it will return a placeholder object which will raise an
    exception only if used.

    Args:
        module (str): The name of the module in which the symbol is defined.
        symbol (str): The name of the symbol to import.
        msg (str, optional): An error message to be displayed if this symbol is used
            after a failed import. Defaults to None.
        alt (object, optional): An object to be used in place of the given symbol if it fails
            to import. Defaults to None.
        fallback_module (str, optional): Alternative name of the model in which the symbol is defined.
            The function will first try to import using the `module` value and if that fails
            will also try the `fallback_module`. Defaults to None.

    Returns:
        tuple: A tuple containing two elements. The first element is the imported symbol,
        the given alternate, or a class derived from UnavailableMeta. The second element
        is a boolean indicating whether the intended import was successful.
    """
    try:
        imported_module = importlib.import_module(module)
        return getattr(imported_module, symbol), True
    except ImportError:
        exception_text = traceback.format_exc()
        logger.debug(f"Import of {module} failed with: {exception_text}")
    except AttributeError:
        # if there is a fallback module try it.
        if fallback_module is not None:
            return safe_import_from(fallback_module, symbol, msg=msg, alt=alt, fallback_module=None)
        exception_text = traceback.format_exc()
        logger.info(f"Import of {symbol} from {module} failed with: {exception_text}")
    except Exception:
        exception_text = traceback.format_exc()
        raise
    if msg is None:
        msg = f"{module}.{symbol} could not be imported"
    if alt is None:
        return UnavailableMeta(symbol, (), {"_msg": msg}), False
    else:
        return alt, False


def gpu_only_import(module, *, alt=None) -> Tuple[object, bool]:
    """A function used to import modules required only in GPU installs.

    This function will attempt to import a module with the given name.
    This function will attempt to import a symbol with the given name from
    the given module, but it will not throw an ImportError if the symbol is not
    found. Instead, it will return a placeholder object which will raise an
    exception only if used with instructions on installing a GPU build.

    Args:
        module (str): The name of the module to import.
        alt (object, optional): A module to be used in place of the given module if it
            fails to import in a non-GPU-enabled install. Defaults to None.

    Returns:
        tuple: A tuple containing two elements. The first element is the imported module,
        the given alternate, or a class derived from UnavailableMeta. The second element
        is a boolean indicating whether the intended import was successful.
    """

    return safe_import(
        module,
        msg=f"{module} is not enabled in non GPU-enabled installations or environemnts. {GPU_INSTALL_STRING}",
        alt=alt,
    )


def gpu_only_import_from(module, symbol, *, alt=None) -> Tuple[object, bool]:
    """A function used to import symbols required only in GPU installs.

    This function will attempt to import a module with the given name.
    This function will attempt to import a symbol with the given name from
    the given module, but it will not throw an ImportError if the symbol is not
    found. Instead, it will return a placeholder object which will raise an
    exception only if used with instructions on installing a GPU build.

    Args:
        module (str): The name of the module to import.
        symbol (str): The name of the symbol to import.
        alt (object, optional): An object to be used in place of the given symbol if it fails
            to import in a non-GPU-enabled install. Defaults to None.

    Returns:
        tuple: A tuple containing two elements. The first element is the imported symbol,
        the given alternate, or a class derived from UnavailableMeta. The second element
        is a boolean indicating whether the intended import was successful.
    """
    return safe_import_from(
        module,
        symbol,
        msg=f"{module}.{symbol} is not enabled in non GPU-enabled installations or environments. {GPU_INSTALL_STRING}",
        alt=alt,
    )


def get_torch_version():
    """Returns the installed PyTorch version as a packaging.version.Version object.

    Handles potential exceptions during version parsing, returning a dummy version
    ("0.0.0") if parsing fails (e.g., during documentation builds where torch
    might not be fully imported or available).

    Returns:
        packaging.version.Version: The parsed PyTorch version, or Version("0.0.0") on error.
    """
    try:
        _torch_version = PkgVersion(torch.__version__)
    except Exception:
        # This is a WAR for building docs, where torch is not actually imported
        _torch_version = PkgVersion("0.0.0")
    return _torch_version


def is_torch_min_version(version, check_equality=True):
    """Check if minimum version of `torch` is installed."""
    if check_equality:
        return get_torch_version() >= PkgVersion(version)
    return get_torch_version() > PkgVersion(version)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch

from megatron.bridge.models.nemotronh import Nemotron3NanoProvider
from megatron.bridge.peft.base import PEFT
from megatron.bridge.peft.lora import LoRA
from megatron.bridge.recipes.common import _peft_common, _pretrain_common, _sft_common
from megatron.bridge.recipes.utils.finetune_utils import default_peft_config
from megatron.bridge.training.comm_overlap import CommOverlapConfig
from megatron.bridge.training.config import ConfigContainer


def nemotron_3_nano_pretrain_config() -> ConfigContainer:
    """Return a pre-training config for Nemotron 3 Nano (30B-A3B MoE).

    This is a MoE (Mixture of Experts) model with the following default parallelism:
    - TP=4, PP=1, EP=8, SP=True
    - DeepEP enabled for MoE token dispatch

    Returns:
        ConfigContainer: Pre-training configuration for Nemotron 3 Nano.
    """
    cfg = _pretrain_common()

    # Model Configuration (MoE)
    cfg.model = Nemotron3NanoProvider(
        tensor_model_parallel_size=4,
        pipeline_model_parallel_size=1,
        pipeline_dtype=torch.bfloat16,
        virtual_pipeline_model_parallel_size=None,
        context_parallel_size=1,
        sequence_parallel=True,
        expert_tensor_parallel_size=1,
        expert_model_parallel_size=8,
        seq_length=8192,
    )

    # Tokenizer (--tokenizer-model)
    cfg.tokenizer.tokenizer_model = "nvidia/NVIDIA-Nemotron-3-Nano-30B-A3B-BF16"

    # Dataset Configuration
    cfg.dataset.seq_length = 8192
    cfg.dataset.blend = None  # Pass the path to the dataset here if not using mock data, along with weight. Ex: (["path/to/data1"], 0.2), [("path/to/data2", 0.8)]
    cfg.dataset.num_workers = 8
    cfg.dataset.mmap_bin_files = False

    # Parallelism Settings (MoE-specific)
    cfg.model.pipeline_model_parallel_layout = None

    # MoE Token Dispatcher Settings
    cfg.model.moe_token_dispatcher_type = "flex"
    cfg.model.moe_flex_dispatcher_backend = "deepep"
    cfg.model.moe_hybridep_num_sms = 16

    # Training Configuration
    cfg.train.train_iters = 39735
    cfg.train.global_batch_size = 3072
    cfg.train.micro_batch_size = 2
    cfg.train.manual_gc = False
    cfg.train.manual_gc_interval = 0

    # Transformer Engine (TE)
    cfg.model.transformer_impl = "transformer_engine"

    # CUDA Graph
    cfg.model.cuda_graph_impl = "none"
    cfg.model.cuda_graph_scope = "full"
    cfg.model.cuda_graph_warmup_steps = 3

    # Kernel Selections
    cfg.model.attention_backend = "fused"
    cfg.model.moe_router_fusion = False
    cfg.model.moe_permute_fusion = True
    cfg.model.moe_grouped_gemm = True
    cfg.model.cross_entropy_loss_fusion = True
    cfg.model.cross_entropy_fusion_impl = "native"

    # Memory Saving (recompute & offloading)
    cfg.model.recompute_granularity = None
    cfg.model.recompute_modules = None
    cfg.model.fine_grained_activation_offloading = False
    cfg.model.offload_modules = None

    # =========================================================================
    # FP8 & MXFP8 (Mixed Precision Settings)
    # =========================================================================
    # Note: mixed_precision="bf16_mixed" is set in _pretrain_common as default
    # FP8 settings (disabled by default, uncomment to enable)
    # cfg.mixed_precision.fp8_recipe = "tensorwise"
    # cfg.mixed_precision.fp8 = None
    # cfg.mixed_precision.fp8_param_gather = False
    # cfg.mixed_precision.reuse_grad_buf_for_mxfp8_param_ag = False
    cfg.model.moe_router_padding_for_fp8 = False

    # Optimizer Precision Settings
    cfg.optimizer.use_precision_aware_optimizer = False
    cfg.optimizer.main_grads_dtype = torch.float32
    cfg.optimizer.main_params_dtype = torch.float32
    cfg.optimizer.exp_avg_dtype = torch.float32
    cfg.optimizer.exp_avg_sq_dtype = torch.float32

    # Optimizer hyperparameters
    cfg.optimizer.lr = 1.6e-3
    cfg.optimizer.weight_decay = 0.1
    cfg.scheduler.min_lr = 1.6e-5
    cfg.scheduler.warmup_iters = 333

    # Communication Overlap
    cfg.comm_overlap = CommOverlapConfig(
        tp_comm_bootstrap_backend="nccl",
        tp_comm_overlap=True,
    )
    cfg.comm_overlap.delay_wgrad_compute = False
    cfg.comm_overlap.overlap_moe_expert_parallel_comm = False
    cfg.model.moe_shared_expert_overlap = False

    # Checkpoint Configuration
    # Paths are set in _pretrain_common by default. Override here if needed:
    # cfg.checkpoint.load = "path/to/load"
    # cfg.checkpoint.save = "path/to/save"
    cfg.checkpoint.save_interval = 200
    cfg.checkpoint.ckpt_assume_constant_structure = True
    cfg.checkpoint.dist_ckpt_strictness = "log_all"

    # DDP Configuration
    cfg.ddp.overlap_grad_reduce = True
    cfg.ddp.overlap_param_gather = True
    cfg.ddp.check_for_nan_in_grad = True
    cfg.ddp.use_distributed_optimizer = True

    # MoE Force Load Balancing
    cfg.model.moe_router_force_load_balancing = False

    cfg.model.init_method_std = 0.0173
    cfg.model.apply_rope_fusion = False
    cfg.model.gradient_accumulation_fusion = True
    cfg.model.use_fused_weighted_squared_relu = True

    return cfg


# =============================================================================
# SFT Config
# =============================================================================


def nemotron_3_nano_sft_config() -> ConfigContainer:
    """Return a full SFT config for Nemotron 3 Nano (30B-A3B MoE).

    Default parallelism: TP=1, PP=1, EP=8, SP=False

    Returns:
        ConfigContainer with all settings pre-configured for Nemotron 3 Nano SFT.
    """
    cfg = _sft_common()

    # Model config - uses Nemotron3NanoProvider
    cfg.model = Nemotron3NanoProvider(
        tensor_model_parallel_size=1,
        pipeline_model_parallel_size=1,
        pipeline_dtype=torch.bfloat16,
        virtual_pipeline_model_parallel_size=None,
        context_parallel_size=1,
        sequence_parallel=False,
        expert_tensor_parallel_size=1,
        expert_model_parallel_size=8,
        apply_rope_fusion=False,
        attention_backend="fused",
        gradient_accumulation_fusion=True,
        init_method_std=0.0173,
        use_fused_weighted_squared_relu=True,
        seq_length=2048,
        calculate_per_token_loss=True,
    )

    # Parallelism settings
    cfg.model.pipeline_model_parallel_layout = None

    # Sequence length
    cfg.model.seq_length = 2048

    # DeePEP settings - set to True to enable DeePEP (enabled by default for Nemotron)
    enable_deepep = True
    if enable_deepep:
        cfg.model.moe_token_dispatcher_type = "flex"
        cfg.model.moe_flex_dispatcher_backend = "deepep"
        cfg.model.moe_shared_expert_overlap = False
    else:
        cfg.model.moe_token_dispatcher_type = "alltoall"
        cfg.model.moe_flex_dispatcher_backend = None
        cfg.model.moe_shared_expert_overlap = True

    cfg.model.moe_hybridep_num_sms = 16

    # TE (Transformer Engine)
    cfg.model.transformer_impl = "transformer_engine"

    # CUDA Graph
    cfg.model.cuda_graph_impl = "none"
    cfg.model.cuda_graph_scope = "full"
    cfg.model.cuda_graph_warmup_steps = 3

    # Kernel selections
    cfg.model.attention_backend = "fused"
    cfg.model.moe_router_fusion = False
    cfg.model.moe_permute_fusion = True
    cfg.model.moe_grouped_gemm = True
    cfg.model.cross_entropy_loss_fusion = True
    cfg.model.cross_entropy_fusion_impl = "native"

    # Memory saving (recompute & offloading)
    cfg.model.recompute_granularity = None
    cfg.model.recompute_modules = None
    cfg.model.fine_grained_activation_offloading = False
    cfg.model.offload_modules = None

    # FP8 & MXFP8 settings
    # Note: mixed_precision="bf16_mixed" is set as default
    # These are defaults for FP8, enable them if using FP8 - FP8 is not enabled by default
    # cfg.mixed_precision.fp8_recipe = "tensorwise"
    # cfg.mixed_precision.fp8 = None
    # cfg.mixed_precision.fp8_param_gather = False
    # cfg.mixed_precision.reuse_grad_buf_for_mxfp8_param_ag = False
    cfg.optimizer.use_precision_aware_optimizer = False
    cfg.optimizer.main_grads_dtype = torch.float32
    cfg.optimizer.main_params_dtype = torch.float32
    cfg.optimizer.exp_avg_dtype = torch.float32
    cfg.optimizer.exp_avg_sq_dtype = torch.float32
    cfg.model.moe_router_padding_for_fp8 = False

    # MoE Force Load Balancing
    cfg.model.moe_router_force_load_balancing = False

    # Training config overrides
    cfg.train.eval_interval = 500

    # Dataset config - packed_sequence=True by default (from _sft_common), seq_length=2048
    # _sft_common already sets seq_length=2048 and packed_sequence=True
    # Adjust pad_seq_to_mult for context parallelism
    if cfg.model.context_parallel_size > 1:
        cfg.dataset.packed_sequence_specs.pad_seq_to_mult = cfg.model.context_parallel_size * 2

    # Optimizer overrides - Nemotron uses specific optimizer settings
    cfg.optimizer.adam_beta2 = 0.95
    cfg.optimizer.adam_eps = 1e-8
    cfg.optimizer.weight_decay = 0.1
    cfg.scheduler.start_weight_decay = 0.1
    cfg.scheduler.end_weight_decay = 0.1
    cfg.scheduler.lr_decay_style = "cosine"

    # Tokenizer
    cfg.tokenizer.tokenizer_model = "nvidia/NVIDIA-Nemotron-3-Nano-30B-A3B-BF16"

    # Checkpoint config overrides
    cfg.checkpoint.save_interval = 200
    cfg.checkpoint.ckpt_format = "torch_dist"
    cfg.checkpoint.dist_ckpt_strictness = "log_all"
    cfg.checkpoint.ckpt_assume_constant_structure = True
    # Uncomment below if using a pretrained checkpoint and provide path to the directory containing pretrained model for finetuning
    # cfg.checkpoint.pretrained_checkpoint = "/path/to/checkpoint"

    # Logger config
    cfg.logger.log_interval = 10
    cfg.logger.log_timers_to_tensorboard = False

    # RNG config - Nemotron uses seed 1234
    cfg.rng.seed = 1234

    # DDP config
    cfg.ddp.check_for_nan_in_grad = True
    cfg.ddp.grad_reduce_in_fp32 = True
    cfg.ddp.overlap_grad_reduce = True
    cfg.ddp.overlap_param_gather = True
    cfg.ddp.use_distributed_optimizer = True

    # Communication overlap settings(default None, can pass CommOverlapConfig for advanced overlap), uncomment to enable
    # cfg.comm_overlap = CommOverlapConfig(
    #     tp_comm_bootstrap_backend="nccl",
    #     tp_comm_overlap=True,
    # )
    # cfg.comm_overlap.delay_wgrad_compute = False
    # cfg.comm_overlap.overlap_moe_expert_parallel_comm = False

    return cfg


# =============================================================================
# PEFT Config
# =============================================================================


def nemotron_3_nano_peft_config(
    peft_scheme: str | PEFT = "lora",
) -> ConfigContainer:
    """Return a PEFT config for Nemotron 3 Nano (30B-A3B MoE).

    Default parallelism: TP=1, PP=1, EP=8, SP=False

    Args:
        peft_scheme: PEFT scheme - "lora", "dora", or a custom PEFT instance.

    Returns:
        ConfigContainer with all settings pre-configured for Nemotron 3 Nano PEFT.
    """
    cfg = _peft_common()

    # Model config - PEFT uses same parallelism as SFT
    cfg.model = Nemotron3NanoProvider(
        tensor_model_parallel_size=1,
        pipeline_model_parallel_size=1,
        pipeline_dtype=torch.bfloat16,
        virtual_pipeline_model_parallel_size=None,
        context_parallel_size=1,
        sequence_parallel=False,
        expert_tensor_parallel_size=1,
        expert_model_parallel_size=8,
        apply_rope_fusion=False,
        attention_backend="fused",
        gradient_accumulation_fusion=True,
        init_method_std=0.0173,
        use_fused_weighted_squared_relu=True,
        seq_length=2048,
        calculate_per_token_loss=True,
    )

    # Parallelism settings
    cfg.model.pipeline_model_parallel_layout = None

    # Sequence length
    cfg.model.seq_length = 2048

    # DeePEP settings - set to True to enable DeePEP (enabled by default for Nemotron)
    enable_deepep = True
    if enable_deepep:
        cfg.model.moe_token_dispatcher_type = "flex"
        cfg.model.moe_flex_dispatcher_backend = "deepep"
        cfg.model.moe_shared_expert_overlap = False
    else:
        cfg.model.moe_token_dispatcher_type = "alltoall"
        cfg.model.moe_flex_dispatcher_backend = None
        cfg.model.moe_shared_expert_overlap = True

    cfg.model.moe_hybridep_num_sms = 16

    # TE (Transformer Engine)
    cfg.model.transformer_impl = "transformer_engine"

    # CUDA Graph
    cfg.model.cuda_graph_impl = "none"
    cfg.model.cuda_graph_scope = "full"
    cfg.model.cuda_graph_warmup_steps = 3

    # Kernel selections
    cfg.model.attention_backend = "fused"
    cfg.model.moe_router_fusion = False
    cfg.model.moe_permute_fusion = True
    cfg.model.moe_grouped_gemm = True
    cfg.model.cross_entropy_loss_fusion = True
    cfg.model.cross_entropy_fusion_impl = "native"

    # Memory saving
    cfg.model.recompute_granularity = None
    cfg.model.recompute_modules = None
    cfg.model.fine_grained_activation_offloading = False
    cfg.model.offload_modules = None

    # FP8 & MXFP8 settings
    # These are defaults for FP8, enable them if using FP8 - FP8 is not enabled by default
    # cfg.mixed_precision.fp8_recipe = "tensorwise"
    # cfg.mixed_precision.fp8 = None
    # cfg.mixed_precision.fp8_param_gather = False
    # cfg.mixed_precision.reuse_grad_buf_for_mxfp8_param_ag = False
    cfg.optimizer.use_precision_aware_optimizer = False
    cfg.optimizer.main_grads_dtype = torch.float32
    cfg.optimizer.main_params_dtype = torch.float32
    cfg.optimizer.exp_avg_dtype = torch.float32
    cfg.optimizer.exp_avg_sq_dtype = torch.float32
    cfg.model.moe_router_padding_for_fp8 = False

    # MoE Force Load Balancing
    cfg.model.moe_router_force_load_balancing = False

    # PEFT config - Nemotron uses Mamba-specific target modules
    mamba_target_modules = ["linear_qkv", "linear_proj", "linear_fc1", "linear_fc2", "in_proj", "out_proj"]
    if isinstance(peft_scheme, str) and peft_scheme.lower() in ["lora", "dora"]:
        cfg.peft = default_peft_config(peft_scheme, target_modules=mamba_target_modules)
    elif isinstance(peft_scheme, PEFT):
        cfg.peft = peft_scheme
    else:
        # Default to LoRA with Mamba target modules
        cfg.peft = LoRA(
            target_modules=mamba_target_modules,
            dim=32,
            alpha=32,
            dropout=0.0,
            dropout_position="pre",
            lora_A_init_method="xavier",
            lora_B_init_method="zero",
        )

    # Training config overrides
    cfg.train.eval_interval = 500

    # Dataset config - packed_sequence=True by default (from _peft_common), seq_length=2048
    # _peft_common already sets seq_length=2048 and packed_sequence=True
    # Adjust pad_seq_to_mult for context parallelism
    if cfg.model.context_parallel_size > 1:
        cfg.dataset.packed_sequence_specs.pad_seq_to_mult = cfg.model.context_parallel_size * 2

    # Optimizer overrides
    cfg.optimizer.adam_beta2 = 0.95
    cfg.optimizer.adam_eps = 1e-8
    cfg.optimizer.weight_decay = 0.1
    cfg.scheduler.start_weight_decay = 0.1
    cfg.scheduler.end_weight_decay = 0.1
    cfg.scheduler.lr_decay_style = "cosine"

    # Tokenizer
    cfg.tokenizer.tokenizer_model = "nvidia/NVIDIA-Nemotron-3-Nano-30B-A3B-BF16"

    # Checkpoint config overrides
    cfg.checkpoint.save_interval = 200
    cfg.checkpoint.ckpt_format = "torch_dist"
    cfg.checkpoint.dist_ckpt_strictness = "log_all"
    cfg.checkpoint.ckpt_assume_constant_structure = True
    # Uncomment below if using a pretrained checkpoint and provide path to the directory containing pretrained model for finetuning
    # cfg.checkpoint.pretrained_checkpoint = "/path/to/checkpoint"

    # Logger config
    cfg.logger.log_interval = 10
    cfg.logger.log_timers_to_tensorboard = False

    # RNG config - Nemotron uses seed 1234
    cfg.rng.seed = 1234

    # DDP config
    cfg.ddp.check_for_nan_in_grad = True
    cfg.ddp.grad_reduce_in_fp32 = True
    cfg.ddp.overlap_grad_reduce = True
    cfg.ddp.overlap_param_gather = True
    cfg.ddp.use_distributed_optimizer = True

    # Communication overlap settings(default None, can pass CommOverlapConfig for advanced overlap), uncomment to enable
    # cfg.comm_overlap = CommOverlapConfig(
    #     tp_comm_bootstrap_backend="nccl",
    #     tp_comm_overlap=True,
    # )
    # cfg.comm_overlap.delay_wgrad_compute = False
    # cfg.comm_overlap.overlap_moe_expert_parallel_comm = False

    return cfg


__all__ = [
    # Pretrain config
    "nemotron_3_nano_pretrain_config",
    # SFT config
    "nemotron_3_nano_sft_config",
    # PEFT config
    "nemotron_3_nano_peft_config",
]



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Type definitions for forward step function definitions.

This module provides comprehensive type definitions for forward step functions used in
Megatron Bridge training. Forward step functions are the core of the training loop,
responsible for performing a single forward pass and returning both the output tensor
and a loss function.

Key Types:
    - ForwardStepCallable: Union of all supported forward step signatures (functions + functors)
    - LossFunction: The partial function returned by forward step functions
    - LossFunctionReturn: The possible return types when calling a loss function

Example Usage:
    >>> from functools import partial
    >>> from megatron.bridge.training.state import GlobalState
    >>>
    >>> def my_forward_step(state: GlobalState, data_iterator, model, return_schedule_plan=False):
    ...     # Access configuration, timers, and training state
    ...     timers = state.timers
    ...     config = state.cfg
    ...
    ...     # Get batch data
    ...     batch = next(data_iterator)
    ...
    ...     # Forward pass with timing
    ...     timers("forward-step").start()
    ...     output_tensor = model(batch['input_ids'])
    ...     timers("forward-step").stop()
    ...
    ...     # Create loss function
    ...     def loss_func(output_tensor):
    ...         loss = compute_loss(output_tensor, batch['labels'])
    ...         num_tokens = batch['labels'].numel()
    ...         loss_reduced = {"lm_loss": loss.detach()}
    ...         return loss, num_tokens, loss_reduced  # ThreeTupleLossReturn
    ...
    ...     return output_tensor, partial(loss_func)
    ...
    >>> # State injection is automatic - no manual binding needed!
    >>> pretrain(config, my_forward_step)
    >>>
    >>> # Functor example (for stateful forward steps)
    >>> class StatefulForwardStep:
    ...     def __init__(self, loss_scale: float = 1.0):
    ...         self.loss_scale = loss_scale
    ...         self.step_count = 0
    ...
    ...     def __call__(self, state: GlobalState, data_iterator, model, return_schedule_plan=False):
    ...         self.step_count += 1
    ...         # ... forward step logic with state tracking ...
    ...         return output_tensor, partial(loss_func)
    ...
    >>> functor = StatefulForwardStep(loss_scale=2.0)
    >>> pretrain(config, functor)
"""

from functools import partial
from typing import Any, Iterable, Protocol, overload

import torch
from megatron.core.models.gpt import GPTModel

from megatron.bridge.training.state import GlobalState


# Loss function return types
LossReduced = dict[str, torch.Tensor]  # Dictionary of loss metrics for logging
TwoTupleLossReturn = tuple[torch.Tensor, LossReduced]  # (loss, loss_reduced) - legacy format
ThreeTupleLossReturn = tuple[
    torch.Tensor, torch.Tensor, LossReduced
]  # (loss, num_tokens, loss_reduced) - per-token loss
InferenceLossReturn = Any  # Any data for inference/non-loss collection (when collect_non_loss_data=True)

# Union of all possible loss function return types
LossFunctionReturn = TwoTupleLossReturn | ThreeTupleLossReturn | InferenceLossReturn

# Type for the loss function that gets called with output_tensor
# This is a partial function that when called returns one of the LossFunctionReturn types
LossFunction = partial[LossFunctionReturn]


class TwoArgForwardStep(Protocol):
    """Protocol for forward step functions with 2 arguments.

    This represents forward step functions that don't need access to GlobalState
    and don't support schedule plan return mode.

    Args:
        data_iterator: Iterator providing training data batches
        model: The GPT model to train

    Returns:
        Tuple of (output_tensor, loss_function)
    """

    def __call__(
        self,
        data_iterator: Iterable,
        model: GPTModel,
    ) -> tuple[torch.Tensor, LossFunction]: ...


class ThreeArgStateForwardStep(Protocol):
    """Protocol for forward step functions with 3 arguments including state.

    This represents forward step functions that need access to GlobalState
    but don't support schedule plan return mode.

    Args:
        state: Global training state containing configuration and runtime objects
        data_iterator: Iterator providing training data batches
        model: The GPT model to train

    Returns:
        Tuple of (output_tensor, loss_function)
    """

    def __call__(
        self,
        state: GlobalState,
        data_iterator: Iterable,
        model: GPTModel,
    ) -> tuple[torch.Tensor, LossFunction]: ...


class ThreeArgForwardStep(Protocol):
    """Protocol for forward step functions with 3 arguments.

    This represents forward step functions that don't need access to GlobalState
    but support schedule plan return mode. These are typically 4-arg functions
    that have had GlobalState pre-bound via functools.partial.

    Args:
        data_iterator: Iterator providing training data batches
        model: The GPT model to train
        return_schedule_plan: Whether to return schedule plan instead of output tensor

    Returns:
        Tuple of (output_tensor, loss_function) or (schedule_plan, loss_function)
    """

    def __call__(
        self,
        data_iterator: Iterable,
        model: GPTModel,
        return_schedule_plan: bool = False,
    ) -> tuple[torch.Tensor, LossFunction]: ...


class FourArgForwardStep(Protocol):
    """Protocol for forward step functions with 4 arguments.

    This represents forward step functions that need access to GlobalState
    and support schedule plan return mode. These are the most complete
    forward step function signatures.

    Args:
        state: Global training state containing configuration and runtime objects
        data_iterator: Iterator providing training data batches
        model: The GPT model to train
        return_schedule_plan: Whether to return schedule plan instead of output tensor

    Returns:
        Tuple of (output_tensor, loss_function) or (schedule_plan, loss_function)
    """

    def __call__(
        self,
        state: GlobalState,
        data_iterator: Iterable,
        model: GPTModel,
        return_schedule_plan: bool = False,
    ) -> tuple[torch.Tensor, LossFunction]: ...


class ForwardStepFunctor(Protocol):
    """Protocol for forward step functors (callable classes).

    This protocol represents classes that implement __call__ with one of the
    supported forward step function signatures. Functors are useful when you
    need to maintain state between forward step calls or implement complex
    forward step logic that benefits from object-oriented design.

    The __call__ method must match one of the supported signatures:
    - (data_iterator, model)
    - (data_iterator, model, return_schedule_plan=False)
             OR (state: GlobalState, data_iterator, model)
    - (state: GlobalState, data_iterator, model, return_schedule_plan=False)

    RECOMMENDED: Use GlobalState type hint for automatic state injection and full access
    to configuration, timers, and training state.

    Examples:
        >>> class MyForwardFunctor:
        ...     def __init__(self, loss_scale: float = 1.0):
        ...         self.loss_scale = loss_scale
        ...         self.call_count = 0
        ...
        ...     def __call__(self, state: GlobalState, data_iterator, model, return_schedule_plan=False):
        ...         self.call_count += 1
        ...         # Access training infrastructure
        ...         timers = state.timers
        ...         config = state.cfg
        ...         # ... forward step logic ...
        ...         return output_tensor, loss_function
        ...
        >>> functor = MyForwardFunctor(loss_scale=2.0)
        >>> pretrain(config, functor)  # State injection is automatic!
    """

    @overload
    def __call__(
        self,
        data_iterator: Iterable,
        model: GPTModel,
    ) -> tuple[torch.Tensor, LossFunction]:
        """2-argument signature: (data_iterator, model)."""
        ...

    @overload
    def __call__(
        self,
        data_iterator: Iterable,
        model: GPTModel,
        return_schedule_plan: bool = False,
    ) -> tuple[torch.Tensor, LossFunction]:
        """3-argument signature: (data_iterator, model, return_schedule_plan)."""
        ...

    @overload
    def __call__(
        self,
        state: GlobalState,
        data_iterator: Iterable,
        model: GPTModel,
    ) -> tuple[torch.Tensor, LossFunction]:
        """3-argument signature with state: (state, data_iterator, model)."""
        ...

    @overload
    def __call__(
        self,
        state: GlobalState,
        data_iterator: Iterable,
        model: GPTModel,
        return_schedule_plan: bool = False,
    ) -> tuple[torch.Tensor, LossFunction]:
        """4-argument signature: (state, data_iterator, model, return_schedule_plan)."""
        ...

    def __call__(self, *args, **kwargs) -> tuple[torch.Tensor, LossFunction]:
        """Execute the forward step.

        The actual implementation must match one of the overloaded signatures above.
        This fallback signature is required by the Protocol but should not be used
        directly - type checkers will use the @overload signatures for validation.
        """
        ...


# Union type for all supported forward step function signatures
ForwardStepFunc = TwoArgForwardStep | ThreeArgStateForwardStep | ThreeArgForwardStep | FourArgForwardStep

# Type alias that includes both functions and functors
ForwardStepCallable = ForwardStepFunc | ForwardStepFunctor



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import fields
from typing import Any, Callable, Dict, Optional, Type, Union

from megatron.core.datasets.blended_megatron_dataset_builder import BlendedMegatronDatasetBuilder
from megatron.core.datasets.blended_megatron_dataset_config import BlendedMegatronDatasetConfig
from megatron.core.datasets.gpt_dataset import GPTDataset, MockGPTDataset
from megatron.core.pipeline_parallel.utils import is_pp_first_stage, is_pp_last_stage
from megatron.core.process_groups_config import ProcessGroupCollection

from megatron.bridge.data.builders.finetuning_dataset import FinetuningDatasetBuilder
from megatron.bridge.data.builders.hf_dataset import HFDatasetBuilder, HFDatasetConfig
from megatron.bridge.data.datasets.fim_dataset import GPTFIMDataset
from megatron.bridge.training.config import (
    DataloaderConfig,
    DatasetBuildContext,
    DatasetProvider,
    FinetuningDatasetConfig,
    GPTDatasetConfig,
    GPTFIMDatasetConfig,
    MockGPTDatasetConfig,
)
from megatron.bridge.training.tokenizers.tokenizer import MegatronTokenizer
from megatron.bridge.utils.common_utils import print_rank_0


def is_dataset_built_on_rank(pg_collection: ProcessGroupCollection) -> bool:
    """Determines whether the dataset should be built on the current rank.

    Datasets are typically built only on the first and last pipeline stages
    and the first tensor parallel rank to save memory and avoid redundancy.

    Returns:
        True if the dataset should be built on the current rank, False otherwise.
    """
    return (is_pp_first_stage(pg_collection.pp) or is_pp_last_stage(pg_collection.pp)) and (
        pg_collection.tp.rank() == 0
    )


def pretrain_train_valid_test_datasets_provider(
    train_val_test_num_samples: list[int], dataset_config: BlendedMegatronDatasetConfig
) -> tuple[GPTDataset, GPTDataset, GPTDataset]:
    """Build pretraining train, validation, and test datasets.

    Uses BlendedMegatronDatasetBuilder to create GPTDataset or MockGPTDataset instances.

    Args:
        train_val_test_num_samples: A list containing the number of samples for
                                    train, validation, and test datasets.
        dataset_config: Configuration object for the blended Megatron dataset.

    Returns:
        A tuple containing the train, validation, and test datasets.
    """

    if dataset_config.mock:
        dataset_type = MockGPTDataset
    elif hasattr(dataset_config, "fim_data"):
        dataset_type = GPTFIMDataset
    else:
        dataset_type = GPTDataset

    print_rank_0("> building train, validation, and test datasets for GPT ...")

    # Build the dataset on all ranks for TP-replicated loading
    train_ds, valid_ds, test_ds = BlendedMegatronDatasetBuilder(
        dataset_type, train_val_test_num_samples, lambda: True, dataset_config
    ).build()

    print_rank_0("> finished creating GPT datasets ...")

    return train_ds, valid_ds, test_ds


def hf_train_valid_test_datasets_provider(
    train_val_test_num_samples: list[int], dataset_config: HFDatasetConfig, tokenizer: MegatronTokenizer
) -> tuple[Any, Any, Any]:
    """Build train, validation, and test datasets from a Hugging Face dataset.

    Uses HFDatasetBuilder to create dataset instances.

    Args:
        train_val_test_num_samples: A list containing the number of samples for
                                    train, validation, and test datasets.
        dataset_config: Configuration object for the Hugging Face dataset.
        tokenizer: The MegatronTokenizer instance.

    Returns:
        A tuple containing the train, validation, and test datasets.
    """
    print_rank_0(
        f"> building train, validation, and test datasets for Huggingface dataset {dataset_config.dataset_name} ..."
    )

    # Get field names from DataloaderConfig to exclude
    dataloader_field_names = {field.name for field in fields(DataloaderConfig)}

    train_ds, valid_ds, test_ds = HFDatasetBuilder(
        tokenizer=tokenizer,
        **{
            field.name: getattr(dataset_config, field.name)
            for field in fields(dataset_config)
            if field.name not in dataloader_field_names
        },
    ).build()

    print_rank_0(f"> finished creating Huggingface dataset {dataset_config.dataset_name} ...")

    return train_ds, valid_ds, test_ds


def finetuning_train_valid_test_datasets_provider(
    train_val_test_num_samples: list[int], dataset_config: FinetuningDatasetConfig, tokenizer: MegatronTokenizer
) -> tuple[Any, Any, Any]:
    """Build finetuning train, validation, and test datasets.

    Uses FinetuningDatasetBuilder to create dataset instances.

    Args:
        train_val_test_num_samples: A list containing the number of samples for
                                    train, validation, and test datasets.
        dataset_config: Configuration object for the finetuning dataset.
        tokenizer: The MegatronTokenizer instance.

    Returns:
        A tuple containing the train, validation, and test datasets.
    """
    print_rank_0(
        f">building train, validation, and test datasets for Finetuning dataset from {dataset_config.dataset_root} ..."
    )

    # Get field names from DataloaderConfig to exclude
    dataloader_field_names = {field.name for field in fields(DataloaderConfig)}

    train_ds, valid_ds, test_ds = FinetuningDatasetBuilder(
        tokenizer=tokenizer,
        **{
            field.name: getattr(dataset_config, field.name)
            for field in fields(dataset_config)
            if field.name not in dataloader_field_names
        },
    ).build()

    print_rank_0(f"> finished creating Finetuning dataset from {dataset_config.dataset_root} ...")

    return train_ds, valid_ds, test_ds


_REGISTRY: Dict[Type[Union[FinetuningDatasetConfig, BlendedMegatronDatasetConfig, HFDatasetConfig]], Callable] = {
    GPTDatasetConfig: pretrain_train_valid_test_datasets_provider,
    GPTFIMDatasetConfig: pretrain_train_valid_test_datasets_provider,
    MockGPTDatasetConfig: pretrain_train_valid_test_datasets_provider,
    HFDatasetConfig: hf_train_valid_test_datasets_provider,
    FinetuningDatasetConfig: finetuning_train_valid_test_datasets_provider,
}


def get_dataset_provider(
    dataset_config: Union[FinetuningDatasetConfig, BlendedMegatronDatasetConfig, HFDatasetConfig, DatasetProvider],
) -> Callable:
    """Get the appropriate dataset provider function based on the config type.

    Supports both registry-based providers and protocol-based providers.

    Args:
        dataset_config: The dataset configuration object.

    Returns:
        The callable dataset provider function corresponding to the config type.
    """
    # Check if config implements the DatasetProvider protocol
    if isinstance(dataset_config, DatasetProvider):

        def protocol_adapter(
            train_val_test_num_samples: list[int],
            config: DatasetProvider,
            tokenizer: Optional[MegatronTokenizer] = None,
        ) -> tuple[Optional[Any], Optional[Any], Optional[Any]]:
            """Adapter function that bridges the protocol interface with the legacy interface."""
            context = DatasetBuildContext(
                train_samples=train_val_test_num_samples[0],
                valid_samples=train_val_test_num_samples[1],
                test_samples=train_val_test_num_samples[2],
                tokenizer=tokenizer,
            )
            return config.build_datasets(context)

        return protocol_adapter

    # Fall back to existing registry
    return _REGISTRY[type(dataset_config)]



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Dict, Mapping, Union

import torch
import torch.nn as nn
from megatron.core import parallel_state
from transformers import Qwen3VLForConditionalGeneration, Qwen3VLMoeForConditionalGeneration

from megatron.bridge.models.conversion.mapping_registry import MegatronMappingRegistry
from megatron.bridge.models.conversion.model_bridge import MegatronModelBridge, WeightConversionTask
from megatron.bridge.models.conversion.param_mapping import (
    AutoMapping,
    GatedMLPMapping,
    QKVMapping,
    ReplicatedMapping,
)
from megatron.bridge.models.hf_pretrained.vlm import PreTrainedVLM
from megatron.bridge.models.qwen_vl.modelling_qwen3_vl.model import Qwen3VLModel
from megatron.bridge.models.qwen_vl.qwen3_vl_provider import Qwen3VLModelProvider, Qwen3VLMoEModelProvider
from megatron.bridge.utils.common_utils import extract_expert_number_from_param


@MegatronModelBridge.register_bridge(
    source=Qwen3VLForConditionalGeneration,
    target=Qwen3VLModel,
    provider=Qwen3VLModelProvider,
    model_type="qwen3_vl",
)
class Qwen3VLBridge(MegatronModelBridge):
    """
    Megatron Bridge for Qwen3-VL Conditional Generation.

    This bridge handles the conversion between HuggingFace Qwen3VLForConditionalGeneration
    and Megatron-Core Qwen3VLModel formats, including weight mappings and
    configuration translation for vision-language models.

    The weight mappings are based on the yan-mbridge implementation which defines:
    - Vision model direct mappings
    - Vision attention layer mappings
    - Vision MLP layer mappings
    - Language model mappings
    - Deepstack visual merger mappings

    Example:
        >>> from megatron.bridge import AutoBridge
        >>> bridge = AutoBridge.from_hf_pretrained("Qwen/Qwen3-VL-8B-Instruct")
        >>> provider = bridge.to_megatron_provider()
    """

    def provider_bridge(self, hf_pretrained: PreTrainedVLM) -> Qwen3VLModelProvider:
        """
        Create a Qwen3VLModelProvider from a HuggingFace pretrained model.

        Args:
            hf_pretrained: HuggingFace pretrained VLM model

        Returns:
            Qwen3VLModelProvider configured with the HF model's parameters
        """
        hf_config = hf_pretrained.config
        text_config = hf_config.text_config

        provider_kwargs = self.hf_config_to_provider_kwargs(text_config)

        vision_config = hf_config.vision_config
        vision_config.torch_dtype = provider_kwargs.get("params_dtype", torch.float32)

        provider = Qwen3VLModelProvider(**provider_kwargs)

        # Qwen3-specific settings
        provider.normalization = "RMSNorm"
        provider.gated_linear_unit = True
        provider.add_qkv_bias = text_config.attention_bias
        provider.add_bias_linear = False
        provider.qk_layernorm = True
        provider.hidden_dropout = 0.0

        # VL-specific overrides
        provider.position_embedding_type = "mrope"
        provider.vision_config = vision_config
        provider.hf_text_config = text_config
        provider.head_dim = text_config.head_dim
        provider.bos_token_id = getattr(text_config, "bos_token_id", 151643)
        provider.eos_token_id = getattr(text_config, "eos_token_id", 151645)
        provider.vision_start_token_id = getattr(hf_config, "vision_start_token_id", 151652)
        provider.vision_end_token_id = getattr(hf_config, "vision_end_token_id", 151653)
        provider.image_token_id = getattr(hf_config, "image_token_id", 151655)
        provider.video_token_id = getattr(hf_config, "video_token_id", 151656)
        provider.mrope_section = text_config.rope_scaling.get("mrope_section", [24, 20, 20])

        return provider

    def mapping_registry(self) -> MegatronMappingRegistry:
        param_mappings = {
            "language_model.embedding.word_embeddings.weight": "model.language_model.embed_tokens.weight",
            "language_model.output_layer.weight": "lm_head.weight",
            "language_model.decoder.final_layernorm.weight": "model.language_model.norm.weight",
            "language_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_weight": "model.language_model.layers.*.input_layernorm.weight",
            "language_model.decoder.layers.*.mlp.linear_fc1.layer_norm_weight": "model.language_model.layers.*.post_attention_layernorm.weight",
            "language_model.decoder.layers.*.self_attention.linear_proj.weight": "model.language_model.layers.*.self_attn.o_proj.weight",
            "language_model.decoder.layers.*.mlp.linear_fc2.weight": "model.language_model.layers.*.mlp.down_proj.weight",
            "language_model.decoder.layers.*.self_attention.q_layernorm.weight": "model.language_model.layers.*.self_attn.q_norm.weight",
            "language_model.decoder.layers.*.self_attention.k_layernorm.weight": "model.language_model.layers.*.self_attn.k_norm.weight",
        }

        mapping_list = []

        # Convert simple 1:1 mappings to AutoMapping objects
        for megatron_param, hf_param in param_mappings.items():
            mapping_list.append(AutoMapping(megatron_param=megatron_param, hf_param=hf_param))

        # Add special mappings that require parameter transformation
        mapping_list.extend(
            [
                # Vision model weights are replicated directly
                # This handles all vision encoder layers, patch embeddings, mergers, etc.
                ReplicatedMapping(
                    megatron_param="vision_model.**",
                    hf_param="model.visual.**",
                ),
                # QKV mapping: Combine separate Q, K, V matrices into single QKV matrix
                QKVMapping(
                    megatron_param="language_model.decoder.layers.*.self_attention.linear_qkv.weight",
                    q="model.language_model.layers.*.self_attn.q_proj.weight",
                    k="model.language_model.layers.*.self_attn.k_proj.weight",
                    v="model.language_model.layers.*.self_attn.v_proj.weight",
                ),
                # QKV bias mapping (if attention_bias is True)
                QKVMapping(
                    megatron_param="language_model.decoder.layers.*.self_attention.linear_qkv.bias",
                    q="model.language_model.layers.*.self_attn.q_proj.bias",
                    k="model.language_model.layers.*.self_attn.k_proj.bias",
                    v="model.language_model.layers.*.self_attn.v_proj.bias",
                ),
                # Gated MLP: Combine gate and up projection matrices into single FC1 matrix
                GatedMLPMapping(
                    megatron_param="language_model.decoder.layers.*.mlp.linear_fc1.weight",
                    gate="model.language_model.layers.*.mlp.gate_proj.weight",
                    up="model.language_model.layers.*.mlp.up_proj.weight",
                ),
            ]
        )

        return MegatronMappingRegistry(*mapping_list)


@MegatronModelBridge.register_bridge(
    source=Qwen3VLMoeForConditionalGeneration,
    target=Qwen3VLModel,
    provider=Qwen3VLMoEModelProvider,
    model_type="qwen3_vl_moe",
)
class Qwen3VLMoEBridge(MegatronModelBridge):
    """
    Megatron Bridge for Qwen3-VL MoE (Mixture of Experts) Conditional Generation.

    This bridge handles the conversion between HuggingFace Qwen3VLMoEForConditionalGeneration
    and Megatron-Core Qwen3VL MoE model formats, including weight mappings and
    configuration translation for vision-language MoE models.

    The weight mappings handle:
    - Vision model weights (same as dense model)
    - Language model MoE layers with expert routing
    - Shared embeddings and output layers
    - QK layernorm specific to Qwen3 architecture

    This bridge works with any Qwen3VL MoE model size and automatically extracts
    the MoE configuration from the HuggingFace model.

    Example:
        >>> from megatron.bridge import AutoBridge
        >>> bridge = AutoBridge.from_hf_pretrained("Qwen/Qwen3-VL-30B-A3B-Instruct")
        >>> provider = bridge.to_megatron_provider()
    """

    def __init__(self):
        super().__init__()
        self.hf_weights_cache: Dict[str, Dict[int, torch.Tensor]] = {}

    def provider_bridge(self, hf_pretrained: PreTrainedVLM) -> Qwen3VLMoEModelProvider:
        hf_config = hf_pretrained.config
        text_config = hf_config.text_config

        provider_kwargs = self.hf_config_to_provider_kwargs(text_config)

        vision_config = hf_config.vision_config
        vision_config.torch_dtype = provider_kwargs.get("params_dtype", torch.float32)

        provider = Qwen3VLMoEModelProvider(**provider_kwargs)

        # Qwen3 MoE-specific settings
        provider.normalization = "RMSNorm"
        provider.gated_linear_unit = True
        provider.add_qkv_bias = text_config.attention_bias
        provider.add_bias_linear = False
        provider.qk_layernorm = True
        provider.hidden_dropout = 0.0

        # MoE specific parameters
        provider.moe_ffn_hidden_size = text_config.moe_intermediate_size
        provider.num_moe_experts = text_config.num_experts
        provider.moe_router_topk = text_config.num_experts_per_tok
        provider.decoder_sparse_step = getattr(text_config, "decoder_sparse_step", 1)
        provider.mlp_only_layers = getattr(text_config, "mlp_only_layers", [])
        provider.moe_grouped_gemm = True
        provider.moe_router_load_balancing_type = "aux_loss"
        provider.moe_aux_loss_coeff = 1e-3
        provider.moe_router_pre_softmax = False
        provider.moe_token_dispatcher_type = "alltoall"
        provider.moe_permute_fusion = True

        # VL-specific overrides
        provider.position_embedding_type = "mrope"
        provider.vision_config = vision_config
        provider.hf_text_config = text_config
        provider.head_dim = getattr(
            text_config, "head_dim", text_config.hidden_size // text_config.num_attention_heads
        )
        provider.bos_token_id = getattr(text_config, "bos_token_id", 151643)
        provider.eos_token_id = getattr(text_config, "eos_token_id", 151645)
        provider.vision_start_token_id = getattr(hf_config, "vision_start_token_id", 151652)
        provider.vision_end_token_id = getattr(hf_config, "vision_end_token_id", 151653)
        provider.image_token_id = getattr(hf_config, "image_token_id", 151655)
        provider.video_token_id = getattr(hf_config, "video_token_id", 151656)
        provider.mrope_section = getattr(text_config, "rope_scaling", {}).get("mrope_section", [24, 20, 20])

        return provider

    def mapping_registry(self) -> MegatronMappingRegistry:
        """
        Return MegatronMappingRegistry containing parameter mappings for MoE models.

        The MoE mappings include:
        1. Standard language model mappings (embeddings, layer norms, output)
        2. Vision model mappings (same as dense model)
        3. QKV mappings with QK layernorm
        4. MoE-specific mappings:
           - Router weights for expert selection
           - Expert MLPs (multiple experts per layer)
           - Pre-MLP layernorm
        5. Deepstack visual merger mappings

        Returns:
            MegatronMappingRegistry with all MoE parameter mappings
        """
        # Language model direct mappings (same as dense model)
        param_mappings = {
            # Embeddings and output layers
            "language_model.embedding.word_embeddings.weight": "model.language_model.embed_tokens.weight",
            "language_model.output_layer.weight": "lm_head.weight",
            "language_model.decoder.final_layernorm.weight": "model.language_model.norm.weight",
            # Layer normalization for attention (TE format - fused into linear)
            "language_model.decoder.layers.*.self_attention.linear_qkv.layer_norm_weight": "model.language_model.layers.*.input_layernorm.weight",
            # Layer normalization (non-TE/quantization format - separate modules)
            "language_model.decoder.layers.*.input_layernorm.weight": "model.language_model.layers.*.input_layernorm.weight",
            # MoE-specific: pre-MLP layernorm (already in non-TE format for MoE)
            "language_model.decoder.layers.*.pre_mlp_layernorm.weight": "model.language_model.layers.*.post_attention_layernorm.weight",
            # Attention output projection
            "language_model.decoder.layers.*.self_attention.linear_proj.weight": "model.language_model.layers.*.self_attn.o_proj.weight",
            # QK layernorm weights (Qwen3 specific)
            "language_model.decoder.layers.*.self_attention.q_layernorm.weight": "model.language_model.layers.*.self_attn.q_norm.weight",
            "language_model.decoder.layers.*.self_attention.k_layernorm.weight": "model.language_model.layers.*.self_attn.k_norm.weight",
            # MoE router weights
            "language_model.decoder.layers.*.mlp.router.weight": "model.language_model.layers.*.mlp.gate.weight",
        }

        mapping_list = []

        # Convert simple 1:1 mappings to AutoMapping objects
        for megatron_param, hf_param in param_mappings.items():
            mapping_list.append(AutoMapping(megatron_param=megatron_param, hf_param=hf_param))

        # Add special mappings that require parameter transformation
        mapping_list.extend(
            [
                # Vision model weights are replicated directly
                ReplicatedMapping(
                    megatron_param="vision_model.**",
                    hf_param="model.visual.**",
                ),
                # QKV mapping: Combine separate Q, K, V matrices
                QKVMapping(
                    megatron_param="language_model.decoder.layers.*.self_attention.linear_qkv.weight",
                    q="model.language_model.layers.*.self_attn.q_proj.weight",
                    k="model.language_model.layers.*.self_attn.k_proj.weight",
                    v="model.language_model.layers.*.self_attn.v_proj.weight",
                ),
                # QKV bias mapping (if attention_bias is True)
                QKVMapping(
                    megatron_param="language_model.decoder.layers.*.self_attention.linear_qkv.bias",
                    q="model.language_model.layers.*.self_attn.q_proj.bias",
                    k="model.language_model.layers.*.self_attn.k_proj.bias",
                    v="model.language_model.layers.*.self_attn.v_proj.bias",
                ),
                ExpertMLPGateUpProjMapping(
                    megatron_param="language_model.decoder.layers.*.mlp.experts.linear_fc1.weight*",
                    hf_param="model.language_model.layers.*.mlp.experts.gate_up_proj",
                ),
                ExpertMLPDownProjMapping(
                    megatron_param="language_model.decoder.layers.*.mlp.experts.linear_fc2.weight*",
                    hf_param="model.language_model.layers.*.mlp.experts.down_proj",
                ),
            ]
        )

        return MegatronMappingRegistry(*mapping_list)

    def maybe_modify_converted_hf_weight(
        self,
        task: WeightConversionTask,
        converted_weights_dict: Dict[str, torch.Tensor],
        hf_state_dict: Mapping[str, torch.Tensor],
    ) -> Dict[str, torch.Tensor]:
        num_experts = self.hf_config.text_config.num_experts
        ep_size = parallel_state.get_expert_model_parallel_world_size()
        experts_per_rank = num_experts // ep_size

        try:
            local_expert_number = extract_expert_number_from_param(task.param_name) % experts_per_rank
        except ValueError:
            # not an expert weight
            return converted_weights_dict

        assert len(converted_weights_dict) == 1, (
            f"There should be only one key in the converted_weights_dict, got keys: {converted_weights_dict.keys()}"
        )
        for key, value in converted_weights_dict.items():
            if key not in self.hf_weights_cache:
                self.hf_weights_cache[key] = {}

            # we end up with ep_size many weights to add to the cache
            # unpack the weights and re-index
            if ep_size == 1:
                self.hf_weights_cache[key][local_expert_number] = value
            else:
                assert value.shape[0] == ep_size
                for i, exp_val in enumerate(value):
                    global_expert_number = local_expert_number + (i * experts_per_rank)
                    self.hf_weights_cache[key][global_expert_number] = exp_val
            if len(self.hf_weights_cache[key]) == num_experts:
                logging.debug(f"All experts are loaded for {key}")
                # all experts are loaded
                if self.hf_weights_cache[key][0].ndim == 3:  # expert 0
                    # gate up
                    merged_hf_gate_weights = torch.cat(
                        [self.hf_weights_cache[key][i][0].unsqueeze(0) for i in range(num_experts)], dim=0
                    )
                    merged_hf_up_weights = torch.cat(
                        [self.hf_weights_cache[key][i][1].unsqueeze(0) for i in range(num_experts)], dim=0
                    )
                    del self.hf_weights_cache[key]
                    return {key: torch.cat([merged_hf_gate_weights, merged_hf_up_weights], dim=-1)}
                elif self.hf_weights_cache[key][0].ndim == 2:  # expert 0
                    # down
                    merged_hf_down_weights = torch.cat(
                        [self.hf_weights_cache[key][i].unsqueeze(0) for i in range(num_experts)], dim=0
                    )
                    del self.hf_weights_cache[key]
                    return {key: merged_hf_down_weights}
                else:
                    raise ValueError(
                        f"Incorrect shape of self.hf_weights_cache[key]: {key} {self.hf_weights_cache[key].shape}"
                    )
            else:
                # not all experts are loaded yet, return empty dict
                logging.debug(f"{len(self.hf_weights_cache[key])}/{num_experts} experts are loaded for {key}")
                return {}


class ExpertMLPDownProjMapping(AutoMapping):
    """Mapping for expert MLP down projection weights between HF and Megatron formats."""

    def hf_to_megatron(self, hf_weights: torch.Tensor, megatron_module: nn.Module) -> torch.Tensor:
        global_expert_number = extract_expert_number_from_param(self.megatron_param)
        # hf_weights: [num_experts, down_in, mlp_out]
        expert_weight = hf_weights[global_expert_number].transpose(0, 1).contiguous()
        return super().hf_to_megatron(expert_weight, megatron_module)

    def megatron_to_hf(self, megatron_weights: torch.Tensor, megatron_module: nn.Module) -> Dict[str, torch.Tensor]:
        # [ep_size, down_in, mlp_out]
        # experts need subsequently merged by maybe_modify_converted_hf_weight
        converted_weights_dict = super().megatron_to_hf(megatron_weights, megatron_module)
        for key in converted_weights_dict:
            converted_weights_dict[key] = converted_weights_dict[key].transpose(-1, -2).contiguous()
        return converted_weights_dict

    def _validate_patterns(self, *args, **kwargs):
        # allow number of wildcards to mismatch in this mapping
        pass


class ExpertMLPGateUpProjMapping(AutoMapping):
    """Mapping for expert MLP gate+up projection using shared GatedMLPMapping logic."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Qwen3-VL MoE expert shards use mismatched wildcard counts; relax validation globally.
        GatedMLPMapping._validate_patterns = lambda *args, **kwargs: None

        # Reuse the generic TP-aware split/gather, but we still handle expert selection
        # and HF<->Megatron transpose at this wrapper layer.
        self._gated_mapping = GatedMLPMapping(
            megatron_param=self.megatron_param,
            gate=f"{self.hf_param}.gate",
            up=f"{self.hf_param}.up",
        )

    def hf_to_megatron(self, hf_weights: Union[torch.Tensor, Dict], megatron_module: nn.Module) -> torch.Tensor:
        global_expert_number = extract_expert_number_from_param(self.megatron_param)
        # hf_weights: [num_experts, mlp_in, fused_gate_up_out]
        expert_weight = hf_weights[global_expert_number].transpose(0, 1).contiguous()

        # HF gate_up_proj is [2 * hidden, hidden]; Megatron expects transposed.
        gate, up = torch.chunk(expert_weight, 2, dim=0)
        return self._gated_mapping.hf_to_megatron({"gate": gate, "up": up}, megatron_module)

    def megatron_to_hf(self, megatron_weights: torch.Tensor, megatron_module: nn.Module) -> Dict[str, torch.Tensor]:
        # Let the shared mapping handle TP/PP/EP gather.
        converted = self._gated_mapping.megatron_to_hf(megatron_weights, megatron_module)
        if not converted:
            return {}

        fused: Dict[str, torch.Tensor] = {}

        # only one pair of gate and up for current group of experts
        for name, tensor in converted.items():
            if name.endswith(".gate"):
                base_name = name[: -len(".gate")]
                # [ep_size, mlp_in, gate_out]
                gate_tensor = tensor.transpose(-1, -2).contiguous()
                # [ep_size, mlp_in, up_out]
                up_tensor = converted.get(f"{base_name}.up").transpose(-1, -2).contiguous()
                assert up_tensor is not None
                # Back to HF fused layout: stack [gate; up] along dim 0.
                # [ep_size, 2, mlp_in, gate_out/up_out]
                fused[base_name] = torch.stack([gate_tensor, up_tensor], dim=0 if up_tensor.ndim == 2 else 1)

        # experts need subsequently merged by maybe_modify_converted_hf_weight
        return fused

    def _validate_patterns(self, *args, **kwargs):
        # allow number of wildcards to mismatch in this mapping
        pass



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from typing import List

from megatron.core.models.gpt import GPTModel as MCoreGPTModel
from transformers.models.qwen2_5_vl.modeling_qwen2_5_vl import Qwen2_5_VLVisionConfig

from megatron.bridge.models.gpt_provider import GPTModelProvider
from megatron.bridge.models.qwen_vl.modeling_qwen25_vl import Qwen25VLModel


@dataclass
class Qwen25VLModelProvider(GPTModelProvider):
    """
    Base model provider for Qwen 2.5 VL Models.
    """

    # VL models shouldn't scatter embeddings across sequence parallel regions because
    # the vision embeddings are going to be inserted into the language embeddings.
    scatter_embedding_sequence_parallel: bool = False
    position_embedding_type: str = "mrope"
    mrope_section: List[int] = field(default_factory=lambda: [16, 24, 24])

    # Vision configuration
    vision_config: Qwen2_5_VLVisionConfig = field(default_factory=Qwen2_5_VLVisionConfig)

    # Token IDs
    bos_token_id: int = 151643
    eos_token_id: int = 151645
    vision_start_token_id: int = 151652
    vision_end_token_id: int = 151653
    vision_token_id: int = 151654
    image_token_id: int = 151655
    video_token_id: int = 151656

    # Freeze options
    freeze_language_model: bool = False
    freeze_vision_model: bool = False
    freeze_vision_projection: bool = False

    def provide(self, pre_process=None, post_process=None, vp_stage=None) -> Qwen25VLModel:
        model = Qwen25VLModel(self, pre_process=pre_process, post_process=post_process, vp_stage=vp_stage)

        # Apply freeze options if any are enabled
        if self.freeze_language_model or self.freeze_vision_model or self.freeze_vision_projection:
            model.freeze(
                freeze_language_model=self.freeze_language_model,
                freeze_vision_model=self.freeze_vision_model,
                freeze_vision_projection=self.freeze_vision_projection,
            )

        return model

    def provide_language_model(self, pre_process=None, post_process=None, vp_stage=None) -> MCoreGPTModel:
        return GPTModelProvider.provide(self, pre_process=pre_process, post_process=post_process, vp_stage=vp_stage)
